import os
import subprocess
import tempfile
import threading
from pathlib import Path
from unittest import mock

import boto3
import moto
import pytest
from dagster_cloud.pex.grpc.server.registry import PexInstallationError, PexS3Registry
from dagster_cloud_cli.core.workspace import PexMetadata

BUCKET = "pex-registry-test-bucket"
PREFIX = "org-storage/test"


def file_content(filepath) -> str:
    with open(filepath, encoding="utf-8") as f:
        return f.read()


@pytest.fixture(scope="module")
def s3_test_bucket():
    with moto.mock_aws():
        with mock.patch.dict(
            os.environ,
            {
                "DAGSTER_CLOUD_SERVERLESS_STORAGE_S3_BUCKET": BUCKET,
                "DAGSTER_CLOUD_SERVERLESS_STORAGE_S3_PREFIX": PREFIX,
            },
        ):
            s3 = boto3.resource("s3", region_name="us-east-1")
            s3.create_bucket(Bucket=BUCKET)
            yield s3.Bucket(BUCKET)


def test_pex_registry_with_bad_pex_files(s3_test_bucket):
    with tempfile.TemporaryDirectory() as tmpdir:
        registry = PexS3Registry(tmpdir)

        with pytest.raises(ValueError, match="prefix"):
            registry.get_pex_executable(PexMetadata(pex_tag="invalid-tag"))

        # add some fake pex files
        for name in ["source-hash1.pex", "deps-hash2.pex", "deps-hash3.pex"]:
            s3_test_bucket.put_object(Key=f"{PREFIX}/pex/{name}", Body=name + "-content")

        with pytest.raises(PexInstallationError):
            registry.get_pex_executable(PexMetadata(pex_tag="files=source-hash1.pex"))

        # no source pex
        with pytest.raises(ValueError, match="no source"):
            registry.get_pex_executable(PexMetadata(pex_tag="files=deps-hash2.pex:deps-hash3.pex"))

        # missing file
        with pytest.raises(Exception, match="404"):
            registry.get_pex_executable(
                PexMetadata(pex_tag="files=deps-nosuchhash.pex:source-hash1.pex")
            )


def test_pex_registry_with_good_pex_files(
    sample_repos_pex_files, s3_test_bucket, venvs_root, monkeypatch
):
    with tempfile.TemporaryDirectory() as tmpdir:
        pex_root = os.path.join(tmpdir, ".pex")
        monkeypatch.setenv("PEX_ROOT", pex_root)
        registry = PexS3Registry(tmpdir)

        # add real pex files
        files = sample_repos_pex_files["repo-1"]
        for filepath in files["all"]:
            s3_test_bucket.put_object(
                Key=f"{PREFIX}/pex/{filepath.name}", Body=filepath.read_bytes()
            )

        pex_executable = registry.get_pex_executable(PexMetadata(pex_tag=files["pex_tag"]))

        # ensure it got unpacked under the venv directory
        assert pex_executable.venv_dirs
        for venv_dir in pex_executable.venv_dirs:
            assert Path(venv_dir).parent == venvs_root
        assert venvs_root in Path(pex_executable.source_path).parents

        # ensure PEX_ROOT cache got cleaned up
        assert not os.path.exists(pex_root) or not os.listdir(pex_root)

        # try to run the pex_executable
        env = {**os.environ, **pex_executable.environ}
        output = subprocess.check_output(
            [pex_executable.source_path, "-m", "dagster", "--version"],
            env=env,
            encoding="utf-8",
        )
        assert "version" in output

        output = subprocess.check_output(
            [pex_executable.source_path, "-c", "import repo_1; print(repo_1.defs);"],
            env={**os.environ, **pex_executable.environ},
            encoding="utf-8",
        )
        assert "Definitions" in output


def test_pex_registry_cleanup_files(sample_repos_pex_files, s3_test_bucket, venvs_root):
    with tempfile.TemporaryDirectory() as tmpdir:
        registry = PexS3Registry(tmpdir)

        # pick two arbitrary repos
        repo_1_files = sample_repos_pex_files["repo-1"]
        repo_2_files = sample_repos_pex_files["repo-with-console-script"]

        # upload all pex files to mock s3
        for filepath in repo_1_files["all"] + repo_2_files["all"]:
            s3_test_bucket.put_object(
                Key=f"{PREFIX}/pex/{filepath.name}", Body=filepath.read_bytes()
            )

        pex_tag1 = repo_1_files["pex_tag"]
        # create a pex tag that re-uses the deps.pex with a new source.pex
        files2 = [filepath.name for filepath in repo_1_files["deps"] + [repo_2_files["source"]]]
        pex_tag2 = "files=" + ":".join(sorted(files2))

        # install both pex executables, should result in 3 venvs (deps venv will be shared)
        pex_executable1 = registry.get_pex_executable(PexMetadata(pex_tag=pex_tag1))
        registry.get_pex_executable(PexMetadata(pex_tag=pex_tag2))
        assert len(os.listdir(venvs_root)) == 3

        # preserve 1 executable, should preserve 2 in-use venvs
        registry.cleanup_unused_files(in_use_pex_metadatas=[PexMetadata(pex_tag=pex_tag1)])
        assert len(os.listdir(venvs_root)) == 2

        # should match the preserved executable
        assert set(os.listdir(venvs_root)) == {
            Path(path).name for path in pex_executable1.venv_dirs
        }

        # re-running cleanup should not add/remove any files
        registry.cleanup_unused_files(in_use_pex_metadatas=[PexMetadata(pex_tag=pex_tag1)])
        assert set(os.listdir(venvs_root)) == {
            Path(path).name for path in pex_executable1.venv_dirs
        }

        # reinstall a previously deleted executable (pex_tag2)
        registry.get_pex_executable(PexMetadata(pex_tag=pex_tag2))
        assert len(os.listdir(venvs_root)) == 3
        # cleanup should not delete the re-added venv
        registry.cleanup_unused_files(
            in_use_pex_metadatas=[PexMetadata(pex_tag=pex_tag1), PexMetadata(pex_tag=pex_tag2)]
        )
        assert len(os.listdir(venvs_root)) == 3

        # cleanup everything
        registry.cleanup_unused_files(in_use_pex_metadatas=[])
        assert len(os.listdir(venvs_root)) == 0

        # test concurrent cleanup and installation
        def install_loop():
            for i in range(5):
                assert registry.get_pex_executable(PexMetadata(pex_tag=pex_tag1))
                assert registry.get_pex_executable(PexMetadata(pex_tag=pex_tag2))

        def cleanup_loop():
            for i in range(1000):
                registry.cleanup_unused_files(in_use_pex_metadatas=[])

        threading.Thread(target=cleanup_loop).start()
        install_loop()


def test_pex_console_scripts(sample_repos_pex_files, s3_test_bucket, venvs_root):
    with tempfile.TemporaryDirectory() as tmpdir:
        registry = PexS3Registry(tmpdir)
        files = sample_repos_pex_files["repo-with-console-script"]
        for filepath in files["all"]:
            s3_test_bucket.put_object(
                Key=f"{PREFIX}/pex/{filepath.name}", Body=filepath.read_bytes()
            )

        pex_executable = registry.get_pex_executable(PexMetadata(pex_tag=files["pex_tag"]))
        assert pex_executable.venv_dirs, "pex did not get unpacked into venvs"
        assert str(venvs_root) in pex_executable.environ["PATH"]

        # try to run the console script from within the pex executable context
        proc = subprocess.run(
            [
                pex_executable.source_path,
                "-c",
                "import subprocess; subprocess.run(['flask', '--version']);",
            ],
            env={**os.environ, **pex_executable.environ},
            encoding="utf-8",
            capture_output=True,
            check=False,
        )
        assert proc.returncode == 0, proc.stderr + proc.stdout
        assert "Flask" in proc.stdout


def test_pex_quickstart_etl(sample_repos_pex_files, s3_test_bucket, venvs_root):
    # this test runs "dagster job list" to ensure the dependencies of quickstart, such as
    # wordcloud, are available
    with tempfile.TemporaryDirectory() as tmpdir:
        registry = PexS3Registry(tmpdir)
        files = sample_repos_pex_files["repo-quickstart-etl"]
        for filepath in files["all"]:
            s3_test_bucket.put_object(
                Key=f"{PREFIX}/pex/{filepath.name}", Body=filepath.read_bytes()
            )

        pex_executable = registry.get_pex_executable(PexMetadata(pex_tag=files["pex_tag"]))
        assert pex_executable.venv_dirs, "pex did not get unpacked into a venv"
        assert str(venvs_root) in pex_executable.environ["PATH"]

        # try to run the console script from within the pex executable context
        proc = subprocess.run(
            [
                "dagster",
                "job",
                "list",
                "--package-name",
                "quickstart_etl",
            ],
            env={**os.environ, **pex_executable.environ},
            encoding="utf-8",
            capture_output=True,
            check=False,
        )
        assert proc.returncode == 0, proc.stderr + proc.stdout

        ### Typical output ###
        # Job: all_assets_job
        # Ops: (Execution Order)
        #     hackernews_topstory_ids
        #     hackernews_topstories
        #     hackernews_topstories_word_cloud
        assert "Job: all_assets_job" in proc.stdout
        assert "hackernews_topstories" in [line.strip() for line in proc.stdout.splitlines()]


def test_pex_requirements_txt(sample_repos_pex_files, s3_test_bucket, venvs_root):
    # test various features of requirements.txt such as local packages and comments
    with tempfile.TemporaryDirectory() as tmpdir:
        registry = PexS3Registry(tmpdir)
        files = sample_repos_pex_files["repo-with-requirements-txt"]
        for filepath in files["all"]:
            s3_test_bucket.put_object(
                Key=f"{PREFIX}/pex/{filepath.name}", Body=filepath.read_bytes()
            )

        pex_executable = registry.get_pex_executable(PexMetadata(pex_tag=files["pex_tag"]))
        assert pex_executable.venv_dirs, "pex did not get unpacked into a venv"
        assert str(venvs_root) in pex_executable.environ["PATH"]

        # try to import the local packages and their dependencies from within the pex executable
        proc = subprocess.run(
            [
                pex_executable.source_path,
                "-c",
                (
                    "import local_package_a, local_package_b, click,"
                    " bottle;print(local_package_a.local_package_a_msg,"
                    " local_package_b.local_package_b_msg);"
                ),
            ],
            env={**os.environ, **pex_executable.environ},
            encoding="utf-8",
            capture_output=True,
            check=False,
        )
        assert proc.returncode == 0, proc.stderr + proc.stdout
        assert "local_package_a" in proc.stdout
        assert "local_package_b" in proc.stdout

        # try to import a dep added using a git+https url
        proc = subprocess.run(
            [
                pex_executable.source_path,
                "-c",
                "import flask;print(flask.__version__);",
            ],
            env={**os.environ, **pex_executable.environ},
            encoding="utf-8",
            capture_output=True,
            check=False,
        )
        assert proc.returncode == 0, proc.stderr + proc.stdout
        assert "2.2.2" in proc.stdout


def test_pex_requirements_txt_hashes(sample_repos_pex_files, s3_test_bucket, venvs_root):
    # test various features of requirements.txt such as local packages and comments
    with tempfile.TemporaryDirectory() as tmpdir:
        registry = PexS3Registry(tmpdir)
        files = sample_repos_pex_files["repo-with-requirements-txt-hashes"]
        for filepath in files["all"]:
            s3_test_bucket.put_object(
                Key=f"{PREFIX}/pex/{filepath.name}", Body=filepath.read_bytes()
            )

        pex_executable = registry.get_pex_executable(PexMetadata(pex_tag=files["pex_tag"]))
        assert pex_executable.venv_dirs, "pex did not get unpacked into a venv"
        assert str(venvs_root) in pex_executable.environ["PATH"]

        # try to import the local packages and their dependencies from within the pex executable
        proc = subprocess.run(
            [pex_executable.source_path, "-c", "import click;"],
            env={**os.environ, **pex_executable.environ},
            encoding="utf-8",
            capture_output=True,
            check=False,
        )
        assert proc.returncode == 0, proc.stderr + proc.stdout



import json
from collections.abc import Sequence
from dataclasses import dataclass
from datetime import datetime, timedelta
from typing import TYPE_CHECKING

from dagster import AssetKey, AssetsDefinition, ScheduleDefinition

from dagster_cloud.dagster_insights.snowflake.snowflake_utils import OPAQUE_ID_SQL_SIGIL

if TYPE_CHECKING:
    from dagster_snowflake import SnowflakeConnection

EVENT_RECORDS_LIMIT = 10000


@dataclass(frozen=True, eq=True)
class AssetMaterializationId:
    run_id: str
    asset_key: AssetKey
    partition: str | None
    step_key: str


QUERY_HISTORY_TIME_PADDING = timedelta(hours=1)  # deal with desynchronized clocks


@dataclass
class SnowflakeInsightsDefinitions:
    assets: Sequence[AssetsDefinition]
    schedule: ScheduleDefinition


def get_cost_data_for_hour(
    snowflake: "SnowflakeConnection",
    start_hour: datetime,
    end_hour: datetime,
) -> list[tuple[str, float, str]]:
    """Given a date range, queries the Snowflake query_history table for all queries that were run
    during that time period and returns a mapping from AssetMaterializationId to the cost of the
    query that produced it, as estimated by Snowflake. The cost is in Snowflake credits.
    """
    opaque_ids_sql = rf"""
    regexp_substr_all(query_text, '{OPAQUE_ID_SQL_SIGIL}\\[\\[\\[(.*?)\\]\\]\\]', 1, 1, 'ce', 1)
    """.strip()

    sql = f"""
WITH
warehouse_sizes AS (
    SELECT 'X-Small' AS warehouse_size, 1 AS credits_per_hour UNION ALL
    SELECT 'Small' AS warehouse_size, 2 AS credits_per_hour UNION ALL
    SELECT 'Medium'  AS warehouse_size, 4 AS credits_per_hour UNION ALL
    SELECT 'Large' AS warehouse_size, 8 AS credits_per_hour UNION ALL
    SELECT 'X-Large' AS warehouse_size, 16 AS credits_per_hour UNION ALL
    SELECT '2X-Large' AS warehouse_size, 32 AS credits_per_hour UNION ALL
    SELECT '3X-Large' AS warehouse_size, 64 AS credits_per_hour UNION ALL
    SELECT '4X-Large' AS warehouse_size, 128 AS credits_per_hour UNION ALL
    SELECT 'missing' as warehouse_size, 0 as credits_per_hour
)
SELECT
    {opaque_ids_sql} as opaque_ids,
    qh.execution_time/(1000*60*60)*wh.credits_per_hour AS query_cost,
    query_id
FROM snowflake.account_usage.query_history AS qh
INNER JOIN warehouse_sizes AS wh
    ON ifnull(qh.warehouse_size, 'missing')=wh.warehouse_size
WHERE
    start_time >= '{start_hour - QUERY_HISTORY_TIME_PADDING}'
    AND start_time <= '{end_hour + QUERY_HISTORY_TIME_PADDING}'
HAVING ARRAY_SIZE(opaque_ids) > 0
        """

    with snowflake.get_connection() as conn:
        with conn.cursor() as cursor:
            result = cursor.execute(sql)
            assert result
            results = result.fetchall()

    costs: list[tuple[str, float, str]] = []

    if not results:
        return []

    total = 0
    for result_opaque_ids, result_cost, query_id in results:
        opaque_ids = json.loads(result_opaque_ids)
        # TODO: is this cost splitting logic correct?
        cost = result_cost / len(opaque_ids)
        total += len(opaque_ids)
        for opaque_id in opaque_ids:
            costs.append((opaque_id, float(cost), query_id))

    return costs



import pytest
from dagster._core.workspace.workspace import DefinitionsSource
from dagster_cloud.workspace.user_code_launcher.utils import get_grpc_server_env
from dagster_cloud_backend.storage.host_cloud.cloud_storage.workspace import (
    DagsterCloudCodeLocationEntry,
)
from dagster_cloud_cli.core.workspace import CodeLocationDeployData, GitMetadata

PORT = 3456


def test_executable_path():
    metadata = CodeLocationDeployData(image="my_image", python_file="my_file.py")
    assert metadata.get_grpc_server_command() == ["dagster", "api", "grpc"]

    with_executable_path = CodeLocationDeployData(
        image="my_image", python_file="my_file.py", executable_path="my_folder/my_executable_path"
    )

    assert with_executable_path.get_grpc_server_command() == [
        "my_folder/my_executable_path",
        "-m",
        "dagster",
        "api",
        "grpc",
    ]


def test_one_python_target():
    with pytest.raises(Exception, match="Must supply exactly one"):
        CodeLocationDeployData(image="my_image")

    with pytest.raises(Exception, match="Must supply exactly one"):
        CodeLocationDeployData(python_file="my_file", package_name="my_package")

    with pytest.raises(Exception, match="Must supply exactly one"):
        CodeLocationDeployData(python_file="my_file", module_name="my_package")


def test_file(host_instance, agent_instance):
    metadata = CodeLocationDeployData(python_file="my_file")
    env = get_grpc_server_env(
        metadata, PORT, "foo_location", agent_instance.ref_for_deployment("staging")
    )
    assert env["DAGSTER_CLI_API_GRPC_PORT"] == str(PORT)
    assert env["DAGSTER_CLI_API_GRPC_HOST"] == "0.0.0.0"
    assert env["DAGSTER_LOCATION_NAME"] == "foo_location"

    assert env.get("DAGSTER_CLI_API_GRPC_LAZY_LOAD_USER_CODE")

    assert not env.get("DAGSTER_CLI_API_GRPC_WORKING_DIRECTORY")
    assert env["DAGSTER_CLI_API_GRPC_PYTHON_FILE"] == "my_file"

    assert not env.get("DAGSTER_CLI_API_GRPC_MODULE_NAME")
    assert not env.get("DAGSTER_CLI_API_GRPC_PACKAGE_NAME")
    assert not env.get("DAGSTER_CLI_API_GRPC_ATTRIBUTE")
    assert not env.get("DAGSTER_CURRENT_IMAGE")

    assert DagsterCloudCodeLocationEntry(
        location_id=1,
        location_name="location",
        code_location_deploy_data=metadata,
        timestamp=0.0,
        definitions_source=DefinitionsSource.CODE_SERVER,
    ).get_display_metadata(host_instance) == {"python_file": "my_file"}


def test_autoload_defs_module_name(host_instance, agent_instance):
    metadata = CodeLocationDeployData(autoload_defs_module_name="autoload_me")
    env = get_grpc_server_env(
        metadata, PORT, "foo_location", agent_instance.ref_for_deployment("staging")
    )
    assert env["DAGSTER_CLI_API_GRPC_PORT"] == str(PORT)
    assert env["DAGSTER_CLI_API_GRPC_HOST"] == "0.0.0.0"
    assert env["DAGSTER_LOCATION_NAME"] == "foo_location"

    assert env.get("DAGSTER_CLI_API_GRPC_LAZY_LOAD_USER_CODE")

    assert env["DAGSTER_CLI_API_GRPC_AUTOLOAD_DEFS_MODULE_NAME"] == "autoload_me"

    assert not env.get("DAGSTER_CLI_API_GRPC_MODULE_NAME")
    assert not env.get("DAGSTER_CLI_API_GRPC_PYTHON_FILE")
    assert not env.get("DAGSTER_CLI_API_GRPC_PACKAGE_NAME")
    assert not env.get("DAGSTER_CLI_API_GRPC_ATTRIBUTE")
    assert not env.get("DAGSTER_CURRENT_IMAGE")

    assert DagsterCloudCodeLocationEntry(
        location_id=1,
        location_name="location",
        code_location_deploy_data=metadata,
        timestamp=0.0,
        definitions_source=DefinitionsSource.CODE_SERVER,
    ).get_display_metadata(host_instance) == {"autoload_defs_module_name": "autoload_me"}


def test_image(host_instance, agent_instance):
    metadata = CodeLocationDeployData(python_file="my_file", image="my_image")
    assert (
        get_grpc_server_env(
            metadata, PORT, "foo_location_name", agent_instance.ref_for_deployment("staging")
        )["DAGSTER_CURRENT_IMAGE"]
        == "my_image"
    )

    assert DagsterCloudCodeLocationEntry(
        location_id=1,
        location_name="loc",
        code_location_deploy_data=metadata,
        timestamp=0.0,
        definitions_source=DefinitionsSource.CODE_SERVER,
    ).get_display_metadata(host_instance) == {
        "image": "my_image",
        "python_file": "my_file",
    }


def test_git_metadata(host_instance, agent_instance):
    metadata = CodeLocationDeployData(
        python_file="my_file",
        image="my_image",
        git_metadata=GitMetadata(commit_hash="abc123", url="www.github.com"),
    )
    assert (
        get_grpc_server_env(
            metadata, PORT, "foo_location", agent_instance.ref_for_deployment("staging")
        )["DAGSTER_CURRENT_IMAGE"]
        == "my_image"
    )

    assert DagsterCloudCodeLocationEntry(
        location_id=1,
        location_name="loc",
        code_location_deploy_data=metadata,
        timestamp=0.0,
        definitions_source=DefinitionsSource.CODE_SERVER,
    ).get_display_metadata(host_instance) == {
        "image": "my_image",
        "python_file": "my_file",
        "url": "www.github.com",
        "commit_hash": "abc123",
    }


def test_package(host_instance, agent_instance):
    metadata = CodeLocationDeployData(package_name="my_package")
    env = get_grpc_server_env(
        metadata, PORT, "foo_location", agent_instance.ref_for_deployment("staging")
    )

    assert not env.get("DAGSTER_CLI_API_GRPC_PYTHON_FILE")
    assert not env.get("DAGSTER_CLI_API_GRPC_MODULE_NAME")
    assert env.get("DAGSTER_CLI_API_GRPC_PACKAGE_NAME") == "my_package"

    assert DagsterCloudCodeLocationEntry(
        location_id=1,
        location_name="loc",
        code_location_deploy_data=metadata,
        timestamp=0.0,
        definitions_source=DefinitionsSource.CODE_SERVER,
    ).get_display_metadata(host_instance) == {
        "package_name": "my_package",
    }


def test_module(host_instance, agent_instance):
    metadata = CodeLocationDeployData(module_name="my_module")
    env = get_grpc_server_env(
        metadata, PORT, "foo_location", agent_instance.ref_for_deployment("staging")
    )

    assert not env.get("DAGSTER_CLI_API_GRPC_PYTHON_FILE")
    assert env.get("DAGSTER_CLI_API_GRPC_MODULE_NAME") == "my_module"
    assert not env.get("DAGSTER_CLI_API_GRPC_PACKAGE_NAME")

    assert DagsterCloudCodeLocationEntry(
        location_id=1,
        location_name="loc",
        code_location_deploy_data=metadata,
        timestamp=0.0,
        definitions_source=DefinitionsSource.CODE_SERVER,
    ).get_display_metadata(host_instance) == {
        "module_name": "my_module",
    }


def test_working_directory(host_instance, agent_instance):
    metadata = CodeLocationDeployData(
        python_file="my_file", working_directory="my_folder/my_directory"
    )
    env = get_grpc_server_env(
        metadata, PORT, "foo_location", agent_instance.ref_for_deployment("staging")
    )
    assert env.get("DAGSTER_CLI_API_GRPC_WORKING_DIRECTORY") == "my_folder/my_directory"

    assert DagsterCloudCodeLocationEntry(
        location_id=1,
        location_name="location",
        code_location_deploy_data=metadata,
        timestamp=0.0,
        definitions_source=DefinitionsSource.CODE_SERVER,
    ).get_display_metadata(host_instance) == {
        "python_file": "my_file",
        "working_directory": "my_folder/my_directory",
    }


def test_attribute(host_instance, agent_instance):
    metadata = CodeLocationDeployData(python_file="my_file", attribute="my_attribute")
    env = get_grpc_server_env(
        metadata, PORT, "foo_location", agent_instance.ref_for_deployment("staging")
    )
    assert env.get("DAGSTER_CLI_API_GRPC_ATTRIBUTE") == "my_attribute"
    assert DagsterCloudCodeLocationEntry(
        location_id=1,
        location_name="location",
        code_location_deploy_data=metadata,
        timestamp=0.0,
        definitions_source=DefinitionsSource.CODE_SERVER,
    ).get_display_metadata(host_instance) == {
        "python_file": "my_file",
        "attribute": "my_attribute",
    }



import os
from unittest import mock

import pytest
from dagster._core.test_utils import environ, instance_for_test
from dagster_cloud import DagsterCloudAgentInstance
from dagster_cloud_backend.storage.host_cloud.cloud_storage.tokens.agent_token import (
    DagsterCloudAgentTokenStorage,
)
from dagster_cloud_backend.storage.host_cloud.cloud_storage.user import DagsterCloudUserStorage
from dagster_cloud_backend.types import DeploymentAgentType
from dagster_cloud_backend.workspace import CloudDagitProcessContext
from dagster_cloud_dagit.server.cloud_dagit import CloudDagitWebserver
from dagster_cloud_test_infra.misc import create_test_host_cloud_instance
from dagster_test.fixtures import *  # noqa: F403
from starlette.testclient import TestClient
from ursula.graphql.implementation.context import UrsulaGraphQLContext
from ursula.server import UrsulaGraphQLServer
from ursula.test_utils import agent_client, ursula_server


@pytest.fixture
def agent_instance(agent_token, ursula_graphql_client):
    with instance_for_test(
        {
            "instance_class": {
                "module": "dagster_cloud",
                "class": "DagsterCloudAgentInstance",
            },
            "user_code_launcher": {
                "module": "dagster_cloud.workspace.user_code_launcher",
                "class": "ProcessUserCodeLauncher",
                "config": {
                    "wait_for_processes": True,
                },
            },
            "dagster_cloud_api": {
                "url": "http://localhost:2874",
                "agent_token": agent_token,
                "deployment": "sandbox",
            },
            "compute_logs": {
                "module": "dagster.core.storage.noop_compute_log_manager",
                "class": "NoOpComputeLogManager",
            },
        }
    ) as instance:
        yield instance


@pytest.fixture
def ursula_asgi_app(host_unscoped_instance):
    ursula_context = UrsulaGraphQLContext(host_unscoped_instance)
    return UrsulaGraphQLServer(ursula_context).create_asgi_app(debug=True)


@pytest.fixture
def ursula_asgi_client(ursula_asgi_app):
    return TestClient(ursula_asgi_app, base_url="http://agent.dagster.cloud")


@pytest.fixture
def dagit_test_client(host_unscoped_instance):
    with environ(
        {
            "SESSION_SECRET_KEY": "fake",
        }
    ):
        context = CloudDagitProcessContext(host_unscoped_instance, version="test")
        yield TestClient(
            CloudDagitWebserver(context).create_asgi_app(debug=True),
            base_url="http://dagster.cloud/",
        )


@pytest.fixture
def agent_instance_local_ursula(agent_token, ursula_asgi_client):
    with (
        mock.patch.object(
            DagsterCloudAgentInstance,
            "client_managed_retries_requests_session",
            new_callable=mock.PropertyMock,
        ) as mock_graphql_requests_session,
        mock.patch.object(
            DagsterCloudAgentInstance,
            "requests_managed_retries_session",
            new_callable=mock.PropertyMock,
        ) as mock_rest_requests_session,
    ):
        # Run ursula locally so that it can use mocked SSM
        with instance_for_test(
            {
                "instance_class": {
                    "module": "dagster_cloud",
                    "class": "DagsterCloudAgentInstance",
                },
                "user_code_launcher": {
                    "module": "dagster_cloud.workspace.user_code_launcher",
                    "class": "ProcessUserCodeLauncher",
                    "config": {
                        "wait_for_processes": True,
                    },
                },
                "dagster_cloud_api": {
                    "url": "http://localhost:2874",
                    "agent_token": agent_token,
                    "deployment": "sandbox",
                },
                "compute_logs": {
                    "module": "dagster.core.storage.noop_compute_log_manager",
                    "class": "NoOpComputeLogManager",
                },
            }
        ) as instance:

            def _mocked_ursula_request(method, url, **kwargs):
                # TestClient uses httpx not requests, not remove kwargs that aren't valid in httpx

                if "verify" in kwargs:
                    kwargs.pop("verify")
                assert not kwargs.get("proxies"), "Proxies not supported in TestClient"

                if "proxies" in kwargs:
                    kwargs.pop("proxies")

                request = ursula_asgi_client.build_request(method, url, **kwargs)
                return ursula_asgi_client.send(request)

            def _mocked_post(url, **kwargs):
                return _mocked_ursula_request("POST", url, **kwargs)

            def _mocked_put(url, **kwargs):
                return _mocked_ursula_request("PUT", url, **kwargs)

            mock_session = mock.MagicMock()
            mock_session.post = _mocked_post
            mock_session.put = _mocked_put
            mock_graphql_requests_session.return_value = mock_session
            mock_rest_requests_session.return_value = mock_session

            yield instance


@pytest.fixture
def server_ttl_agent_instance(agent_token, ursula_graphql_client):
    with instance_for_test(
        {
            "instance_class": {
                "module": "dagster_cloud",
                "class": "DagsterCloudAgentInstance",
            },
            "user_code_launcher": {
                "module": "dagster_cloud.workspace.user_code_launcher",
                "class": "ProcessUserCodeLauncher",
                "config": {
                    "wait_for_processes": True,
                    "server_ttl": {
                        "full_deployments": {
                            "enabled": True,
                            "ttl_seconds": 7200,
                        },
                        "max_servers": 5,
                    },
                },
            },
            "dagster_cloud_api": {
                "url": "http://localhost:2874",
                "agent_token": agent_token,
                "deployment": "sandbox",
            },
            "compute_logs": {
                "module": "dagster.core.storage.noop_compute_log_manager",
                "class": "NoOpComputeLogManager",
            },
        }
    ) as instance:
        yield instance


@pytest.fixture
def branch_deployments_agent_instance(agent_token, ursula_graphql_client):
    with instance_for_test(
        {
            "instance_class": {
                "module": "dagster_cloud",
                "class": "DagsterCloudAgentInstance",
            },
            "user_code_launcher": {
                "module": "dagster_cloud.workspace.user_code_launcher",
                "class": "ProcessUserCodeLauncher",
                "config": {
                    "wait_for_processes": True,
                },
            },
            "dagster_cloud_api": {
                "url": "http://localhost:2874",
                "agent_token": agent_token,
                "branch_deployments": True,
            },
            "compute_logs": {
                "module": "dagster.core.storage.noop_compute_log_manager",
                "class": "NoOpComputeLogManager",
            },
        }
    ) as instance:
        yield instance


@pytest.fixture
def branch_deployments_and_prod_deployments_instance(agent_token, ursula_graphql_client):
    with instance_for_test(
        {
            "instance_class": {
                "module": "dagster_cloud",
                "class": "DagsterCloudAgentInstance",
            },
            "user_code_launcher": {
                "module": "dagster_cloud.workspace.user_code_launcher",
                "class": "ProcessUserCodeLauncher",
                "config": {
                    "wait_for_processes": True,
                },
            },
            "dagster_cloud_api": {
                "url": "http://localhost:2874",
                "agent_token": agent_token,
                "branch_deployments": True,
                "deployments": ["sandbox", "prod"],
            },
            "compute_logs": {
                "module": "dagster.core.storage.noop_compute_log_manager",
                "class": "NoOpComputeLogManager",
            },
        }
    ) as instance:
        yield instance


@pytest.fixture
def serverless_instance(agent_token, ursula_graphql_client):
    with instance_for_test(
        {
            "instance_class": {
                "module": "dagster_cloud",
                "class": "DagsterCloudAgentInstance",
            },
            "user_code_launcher": {
                "module": "dagster_cloud.workspace.user_code_launcher",
                "class": "ProcessUserCodeLauncher",
                "config": {
                    "wait_for_processes": True,
                },
            },
            "dagster_cloud_api": {
                "url": "http://localhost:2874",
                "agent_token": agent_token,
                "branch_deployments": True,  # Make configurable from server
                "all_serverless_deployments": True,
            },
            "compute_logs": {
                "module": "dagster.core.storage.noop_compute_log_manager",
                "class": "NoOpComputeLogManager",
            },
        }
    ) as instance:
        yield instance


@pytest.fixture
def isolated_agents_instance(agent_token, ursula_graphql_client):
    with instance_for_test(
        {
            "instance_class": {
                "module": "dagster_cloud",
                "class": "DagsterCloudAgentInstance",
            },
            "user_code_launcher": {
                "module": "dagster_cloud.workspace.user_code_launcher",
                "class": "ProcessUserCodeLauncher",
            },
            "dagster_cloud_api": {
                "url": "http://localhost:2874",
                "agent_token": agent_token,
                "deployment": "sandbox",
            },
            "compute_logs": {
                "module": "dagster.core.storage.noop_compute_log_manager",
                "class": "NoOpComputeLogManager",
            },
            "agent_replicas": {"enabled": True},
        }
    ) as instance:
        yield instance


@pytest.fixture
def agent_queues_agent_instance(agent_token, ursula_graphql_client):
    with instance_for_test(
        {
            "instance_class": {
                "module": "dagster_cloud",
                "class": "DagsterCloudAgentInstance",
            },
            "user_code_launcher": {
                "module": "dagster_cloud.workspace.user_code_launcher",
                "class": "ProcessUserCodeLauncher",
            },
            "dagster_cloud_api": {
                "url": "http://localhost:2874",
                "agent_token": agent_token,
                "deployment": "sandbox",
            },
            "compute_logs": {
                "module": "dagster.core.storage.noop_compute_log_manager",
                "class": "NoOpComputeLogManager",
            },
            "agent_queues": {
                "include_default_queue": False,
                "additional_queues": ["foo", "bar"],
            },
        }
    ) as instance:
        yield instance


@pytest.fixture
def legacy_snapshot_upload_agent_instance(agent_token, ursula_graphql_client):
    with instance_for_test(
        {
            "instance_class": {
                "module": "dagster_cloud",
                "class": "DagsterCloudAgentInstance",
            },
            "user_code_launcher": {
                "module": "dagster_cloud.workspace.user_code_launcher",
                "class": "ProcessUserCodeLauncher",
                "config": {
                    "wait_for_processes": True,
                    "direct_snapshot_uploads": False,  # opted out of direct
                },
            },
            "dagster_cloud_api": {
                "url": "http://localhost:2874",
                "agent_token": agent_token,
                "deployment": "sandbox",
            },
            "compute_logs": {
                "module": "dagster.core.storage.noop_compute_log_manager",
                "class": "NoOpComputeLogManager",
            },
        }
    ) as instance:
        yield instance


@pytest.fixture(scope="module")
def host_unscoped_instance(
    postgres_conn_from_docker_service,
    postgres_shard1_conn_from_docker_service,
    s3_bucket,
    redis,
    reporting_postgres_conn_from_docker_service,
    xregion_postgres_conn_from_docker_service,
    agent_token,
):
    with create_test_host_cloud_instance(
        postgres_conn_from_docker_service,
        s3_bucket,
        redis,
        reporting_postgres_conn_from_docker_service,
        xregion_postgres_conn_from_docker_service,
        postgres_shard_dbs=[
            postgres_shard1_conn_from_docker_service,
        ],
    ) as instance:
        organization = instance.cloud_storage.create_organization_object("acme")
        instance.cloud_storage.create_organization_subdomain("acme", organization.organization_id)
        instance.check_loaded()

        with instance.for_organization(organization) as org_instance:
            org_instance.cloud_storage.create_deployment("sandbox")
            org_instance.cloud_storage.create_deployment("prod")
            org_instance.cloud_storage.create_deployment(
                "serverless",
                agent_type=DeploymentAgentType.SERVERLESS,
            )

            test_user = DagsterCloudUserStorage.add_to_organization(org_instance, "test@acme.com")
            DagsterCloudAgentTokenStorage.create(
                org_instance,
                created_by=test_user.user_id,
                token_value=agent_token,
            )

        yield instance


@pytest.fixture
def host_instance(host_unscoped_instance):
    with host_unscoped_instance.load_for_deployment_by_non_unique_name(
        "acme", "sandbox"
    ) as scoped_instance:
        yield scoped_instance


@pytest.fixture(scope="module")
def ursula_graphql_client(agent_token, host_unscoped_instance):
    os.environ["URSULA_USE_CLIENT_AGENT_HEARTBEAT_TIME"] = "true"
    with ursula_server() as _server_process:
        with agent_client(agent_token) as client:
            yield client


@pytest.fixture
def allowed_locations_agent_instance(agent_token, ursula_graphql_client):
    with instance_for_test(
        {
            "instance_class": {
                "module": "dagster_cloud",
                "class": "DagsterCloudAgentInstance",
            },
            "user_code_launcher": {
                "module": "dagster_cloud.workspace.user_code_launcher",
                "class": "ProcessUserCodeLauncher",
            },
            "dagster_cloud_api": {
                "url": "http://localhost:2874",
                "agent_token": agent_token,
                "deployment": "sandbox",
            },
            "compute_logs": {
                "module": "dagster.core.storage.noop_compute_log_manager",
                "class": "NoOpComputeLogManager",
            },
            "allowed_full_deployment_locations": {
                "sandbox": ["allowed-location"],
            },
        }
    ) as instance:
        yield instance


@pytest.fixture
def allowed_branch_locations_agent_instance(agent_token, ursula_graphql_client):
    with instance_for_test(
        {
            "instance_class": {
                "module": "dagster_cloud",
                "class": "DagsterCloudAgentInstance",
            },
            "user_code_launcher": {
                "module": "dagster_cloud.workspace.user_code_launcher",
                "class": "ProcessUserCodeLauncher",
            },
            "dagster_cloud_api": {
                "url": "http://localhost:2874",
                "agent_token": agent_token,
                "branch_deployments": True,
            },
            "compute_logs": {
                "module": "dagster.core.storage.noop_compute_log_manager",
                "class": "NoOpComputeLogManager",
            },
            "allowed_branch_deployment_locations": ["allowed-branch-location"],
        }
    ) as instance:
        yield instance



import socket
from contextlib import ExitStack
from typing import TYPE_CHECKING, cast
from unittest import mock

import pytest
from dagster import DagsterInstance
from dagster._core.errors import DagsterInvalidConfigError, DagsterInvariantViolationError
from dagster._core.test_utils import environ, instance_for_test
from dagster_cloud.auth.constants import get_hardcoded_test_user_token
from dagster_cloud.storage.client import DEFAULT_TIMEOUT
from dagster_cloud.workspace.user_code_launcher import DEFAULT_SERVER_TTL_SECONDS
from dagster_cloud.workspace.user_code_launcher.utils import get_instance_ref_for_user_code
from dagster_cloud_cli.core.graphql_client import DEFAULT_RETRIES

from dagster_cloud_tests import gen_agent_instance

if TYPE_CHECKING:
    from dagster_cloud.instance import DagsterCloudAgentInstance

WORKSPACE_ENTRIES_QUERY = """
    query WorkspaceEntries {
        workspace {
            workspaceEntries {
                locationName
                serializedDeploymentMetadata
                hasOutdatedData
            }
        }
    }
"""


def test_instance_graphql_connect_failure(agent_instance):
    # Can create an instance even if there is no graphql server to connect to
    # but it throws an error when you make a graphql query

    with pytest.raises(Exception):
        agent_instance.graphql_client.execute(WORKSPACE_ENTRIES_QUERY)


def test_instance_url(agent_instance, dagster_cloud_url):
    assert agent_instance.dagster_cloud_url == dagster_cloud_url
    assert agent_instance.dagster_cloud_graphql_url == f"{dagster_cloud_url}/graphql"


def test_instance_dagster_cloud_env_vars():
    with environ(
        {
            "CLOUD_URL": "my_url",
            "CLOUD_AGENT_TOKEN": "my_agent_token",
            "CLOUD_AGENT_RETRIES": "3",
        }
    ):
        with instance_for_test(
            {
                "instance_class": {
                    "module": "dagster_cloud",
                    "class": "DagsterCloudAgentInstance",
                },
                "user_code_launcher": {
                    "module": "dagster_cloud.workspace.user_code_launcher",
                    "class": "ProcessUserCodeLauncher",
                },
                "dagster_cloud_api": {
                    "url": {"env": "CLOUD_URL"},
                    "agent_token": {"env": "CLOUD_AGENT_TOKEN"},
                    "deployment": "scalar_deployment_name",
                    "retries": {"env": "CLOUD_AGENT_RETRIES"},
                    "method": "POST",
                },
            }
        ) as test_instance:
            assert sorted(test_instance.dagster_cloud_api_env_vars) == sorted(  # pyright: ignore[reportAttributeAccessIssue]
                ["CLOUD_URL", "CLOUD_AGENT_TOKEN", "CLOUD_AGENT_RETRIES"]
            )


def test_default_timeout(agent_instance):
    assert agent_instance.dagster_cloud_api_timeout == DEFAULT_TIMEOUT
    assert agent_instance.graphql_client.timeout == DEFAULT_TIMEOUT


def test_custom_timeout(dagster_cloud_url):
    with gen_agent_instance(dagster_cloud_url, token="token", timeout=15) as custom_instance:
        assert custom_instance.dagster_cloud_api_timeout == 15  # pyright: ignore[reportAttributeAccessIssue]
        assert custom_instance.graphql_client.timeout == 15  # pyright: ignore[reportAttributeAccessIssue]


def test_url_with_new_token_set():
    with gen_agent_instance(token="agent:foo:abcde", url=None) as instance:
        assert instance.dagster_cloud_url == "https://foo.agent.dagster.cloud"  # pyright: ignore[reportAttributeAccessIssue]
        assert instance.dagit_url == "https://foo.dagster.cloud/"  # pyright: ignore[reportAttributeAccessIssue]

    with gen_agent_instance(token="agent:foo:abcde", url=None, deployment="staging") as instance:
        assert instance.dagit_url == "https://foo.dagster.cloud/staging/"  # pyright: ignore[reportAttributeAccessIssue]


def test_url_with_old_token_set(dagster_cloud_url):
    with gen_agent_instance(token="agent_abcde", url=dagster_cloud_url) as instance:
        assert instance.dagster_cloud_url == dagster_cloud_url  # pyright: ignore[reportAttributeAccessIssue]

        with pytest.raises(Exception):
            instance.dagit_url  # noqa: B018  # pyright: ignore[reportAttributeAccessIssue]


def test_url_with_both_set(dagster_cloud_url):
    with gen_agent_instance(token="agent:foo:abcde", url=dagster_cloud_url) as instance:
        assert instance.dagster_cloud_url == dagster_cloud_url  # pyright: ignore[reportAttributeAccessIssue]
        assert instance.dagit_url == "https://foo.dagster.cloud/"  # pyright: ignore[reportAttributeAccessIssue]


def test_url_with_old_token_and_no_url():
    with pytest.raises(
        DagsterInvariantViolationError,
        match=(
            r"Could not derive Dagster Cloud URL from agent token. Create a new agent token or set"
            " the `url` field under `dagster_cloud_api` in your `dagster.yaml`."
        ),
    ):
        with gen_agent_instance(token="agent_abcde", url=None):
            pass


def test_url_with_no_token_and_no_url():
    with pytest.raises(DagsterInvalidConfigError):
        with gen_agent_instance(token=None, url=None):
            pass


def test_url_with_user_token_set():
    user_token = get_hardcoded_test_user_token("arrakis", "paul_atreides")
    with pytest.raises(
        DagsterInvariantViolationError,
        match=(
            r"Agent was configured with a user token, but agents can only authenticate with "
            "Dagster Cloud when configured with an agent token."
        ),
    ):
        with gen_agent_instance(token=user_token, url=None):
            pass


# Verify that two graphql clients share a session
def test_shared_session(agent_instance):
    client1 = agent_instance.create_graphql_client()

    client2 = agent_instance.create_graphql_client()

    assert client1.session == client2.session


def test_code_server_metrics():
    with instance_for_test(
        {
            "instance_class": {
                "module": "dagster_cloud",
                "class": "DagsterCloudAgentInstance",
            },
            "user_code_launcher": {
                "module": "dagster_cloud.workspace.user_code_launcher",
                "class": "ProcessUserCodeLauncher",
            },
            "dagster_cloud_api": {"agent_token": "my_agent_token", "url": "my_url"},
        }
    ) as instance:
        instance = cast("DagsterCloudAgentInstance", instance)
        assert not instance.user_code_launcher.code_server_metrics_enabled

    with instance_for_test(
        {
            "instance_class": {
                "module": "dagster_cloud",
                "class": "DagsterCloudAgentInstance",
            },
            "user_code_launcher": {
                "module": "dagster_cloud.workspace.user_code_launcher",
                "class": "ProcessUserCodeLauncher",
                "config": {
                    "code_server_metrics": {"enabled": True},
                },
            },
            "dagster_cloud_api": {"agent_token": "my_agent_token", "url": "my_url"},
        }
    ) as instance:
        instance = cast("DagsterCloudAgentInstance", instance)
        assert instance.user_code_launcher.code_server_metrics_enabled


def test_isolated_agents():
    with instance_for_test(
        {
            "instance_class": {
                "module": "dagster_cloud",
                "class": "DagsterCloudAgentInstance",
            },
            "user_code_launcher": {
                "module": "dagster_cloud.workspace.user_code_launcher",
                "class": "ProcessUserCodeLauncher",
            },
            "dagster_cloud_api": {"agent_token": "my_agent_token", "url": "my_url"},
        }
    ) as instance:
        assert not instance.is_using_isolated_agents  # pyright: ignore[reportAttributeAccessIssue]

    with instance_for_test(
        {
            "instance_class": {
                "module": "dagster_cloud",
                "class": "DagsterCloudAgentInstance",
            },
            "user_code_launcher": {
                "module": "dagster_cloud.workspace.user_code_launcher",
                "class": "ProcessUserCodeLauncher",
            },
            "dagster_cloud_api": {"agent_token": "my_agent_token", "url": "my_url"},
            "isolated_agents": {"enabled": False},
        }
    ) as instance:
        assert not instance.is_using_isolated_agents  # pyright: ignore[reportAttributeAccessIssue]

    with instance_for_test(
        {
            "instance_class": {
                "module": "dagster_cloud",
                "class": "DagsterCloudAgentInstance",
            },
            "user_code_launcher": {
                "module": "dagster_cloud.workspace.user_code_launcher",
                "class": "ProcessUserCodeLauncher",
            },
            "dagster_cloud_api": {"agent_token": "my_agent_token", "url": "my_url"},
            "isolated_agents": {"enabled": True},
        }
    ) as instance:
        assert instance.is_using_isolated_agents  # pyright: ignore[reportAttributeAccessIssue]

    # Backcompat with agent_replicas as config

    with instance_for_test(
        {
            "instance_class": {
                "module": "dagster_cloud",
                "class": "DagsterCloudAgentInstance",
            },
            "user_code_launcher": {
                "module": "dagster_cloud.workspace.user_code_launcher",
                "class": "ProcessUserCodeLauncher",
            },
            "dagster_cloud_api": {"agent_token": "my_agent_token", "url": "my_url"},
            "agent_replicas": {"enabled": False},
        }
    ) as instance:
        assert not instance.is_using_isolated_agents  # pyright: ignore[reportAttributeAccessIssue]

    with instance_for_test(
        {
            "instance_class": {
                "module": "dagster_cloud",
                "class": "DagsterCloudAgentInstance",
            },
            "user_code_launcher": {
                "module": "dagster_cloud.workspace.user_code_launcher",
                "class": "ProcessUserCodeLauncher",
            },
            "dagster_cloud_api": {"agent_token": "my_agent_token", "url": "my_url"},
            "agent_replicas": {"enabled": True},
        }
    ) as instance:
        assert instance.is_using_isolated_agents  # pyright: ignore[reportAttributeAccessIssue]


def test_instance_with_python_logs(dagster_cloud_url):
    managed_python_loggers = ["root", "some_logger"]
    dagster_handler_config = {
        "handlers": {
            "handlerOne": {
                "class": "logging.StreamHandler",
                "level": "INFO",
                "stream": "ext://sys.stdout",
            },
            "handlerTwo": {
                "class": "logging.StreamHandler",
                "level": "ERROR",
                "stream": "ext://sys.stdout",
            },
        },
    }
    with gen_agent_instance(
        url=dagster_cloud_url,
        token="token",
        python_logs={
            "managed_python_loggers": managed_python_loggers,
            "python_log_level": "INFO",
            "dagster_handler_config": dagster_handler_config,
        },
    ) as instance:
        assert instance.python_log_level == "INFO"
        assert instance.managed_python_loggers == managed_python_loggers
        assert (
            instance.get_settings("python_logs").get("dagster_handler_config")
            == dagster_handler_config
        )


def test_instance_deployment_rescope():
    with ExitStack() as stack:
        unscoped_instance = stack.enter_context(gen_agent_instance(token="agent:foo:abcde"))

        assert unscoped_instance.dagit_url == "https://foo.dagster.cloud/"  # pyright: ignore[reportAttributeAccessIssue]

        scoped_instance = stack.enter_context(
            DagsterInstance.from_ref(unscoped_instance.ref_for_deployment("bar"))  # pyright: ignore[reportAttributeAccessIssue]
        )
        assert scoped_instance.dagit_url == "https://foo.dagster.cloud/bar/"  # pyright: ignore[reportAttributeAccessIssue]

        rescoped_instance = stack.enter_context(
            DagsterInstance.from_ref(scoped_instance.ref_for_deployment("baz"))  # pyright: ignore[reportAttributeAccessIssue]
        )
        assert rescoped_instance.dagit_url == "https://foo.dagster.cloud/baz/"  # pyright: ignore[reportAttributeAccessIssue]

        multi_deployment_scoped_instance = stack.enter_context(
            gen_agent_instance(
                token="agent:foo:abcde",
                dagster_cloud_api_config={"deployments": ["prod", "staging"]},
            )
        )

        rescoped_instance = stack.enter_context(
            DagsterInstance.from_ref(multi_deployment_scoped_instance.ref_for_deployment("staging"))  # pyright: ignore[reportAttributeAccessIssue]
        )
        assert rescoped_instance.dagit_url == "https://foo.dagster.cloud/staging/"  # pyright: ignore[reportAttributeAccessIssue]


def test_instance_branch_deployments(dagster_cloud_url):
    with gen_agent_instance(
        url=dagster_cloud_url, token="token", dagster_cloud_api_config={"branch_deployments": True}
    ) as instance:
        assert instance.includes_branch_deployments  # pyright: ignore[reportAttributeAccessIssue]
        assert (
            instance.user_code_launcher.branch_deployment_ttl_seconds == DEFAULT_SERVER_TTL_SECONDS  # pyright: ignore[reportAttributeAccessIssue]
        )

    with gen_agent_instance(
        url=dagster_cloud_url, token="token", dagster_cloud_api_config={"branch_deployments": False}
    ) as instance:
        assert not instance.includes_branch_deployments  # pyright: ignore[reportAttributeAccessIssue]
        assert (
            instance.user_code_launcher.branch_deployment_ttl_seconds == DEFAULT_SERVER_TTL_SECONDS  # pyright: ignore[reportAttributeAccessIssue]
        )

    with gen_agent_instance(
        url=dagster_cloud_url,
        token="token",
    ) as instance:
        assert not instance.includes_branch_deployments  # pyright: ignore[reportAttributeAccessIssue]
        assert (
            instance.user_code_launcher.branch_deployment_ttl_seconds == DEFAULT_SERVER_TTL_SECONDS  # pyright: ignore[reportAttributeAccessIssue]
        )

    with gen_agent_instance(
        url=dagster_cloud_url,
        token="token",
        user_code_launcher_config={
            "server_ttl": {
                "branch_deployments": {
                    "ttl_seconds": 12345,
                }
            }
        },
    ) as instance:
        assert not instance.includes_branch_deployments  # pyright: ignore[reportAttributeAccessIssue]
        assert instance.user_code_launcher.branch_deployment_ttl_seconds == 12345  # pyright: ignore[reportAttributeAccessIssue]


def test_instance_server_ttl(dagster_cloud_url):
    with gen_agent_instance(
        url=dagster_cloud_url,
        token="token",
    ) as instance:
        assert not instance.user_code_launcher.server_ttl_enabled_for_full_deployments  # pyright: ignore[reportAttributeAccessIssue]
        assert instance.user_code_launcher.full_deployment_ttl_seconds == DEFAULT_SERVER_TTL_SECONDS  # pyright: ignore[reportAttributeAccessIssue]

    with gen_agent_instance(
        url=dagster_cloud_url,
        token="token",
        user_code_launcher_config={
            "server_ttl": {
                "full_deployments": {
                    "enabled": True,
                }
            }
        },
    ) as instance:
        assert instance.user_code_launcher.server_ttl_enabled_for_full_deployments  # pyright: ignore[reportAttributeAccessIssue]
        assert instance.user_code_launcher.full_deployment_ttl_seconds == DEFAULT_SERVER_TTL_SECONDS  # pyright: ignore[reportAttributeAccessIssue]

    with gen_agent_instance(
        url=dagster_cloud_url,
        token="token",
        user_code_launcher_config={
            "server_ttl": {
                "full_deployments": {
                    "enabled": False,
                    "ttl_seconds": 12345,
                }
            }
        },
    ) as instance:
        assert not instance.user_code_launcher.server_ttl_enabled_for_full_deployments  # pyright: ignore[reportAttributeAccessIssue]
        assert instance.user_code_launcher.full_deployment_ttl_seconds == 12345  # pyright: ignore[reportAttributeAccessIssue]


def test_instance_deployment_names(dagster_cloud_url):
    with gen_agent_instance(url=dagster_cloud_url, token="token", deployment="sandbox") as instance:
        assert instance.deployment_names == ["sandbox"]  # pyright: ignore[reportAttributeAccessIssue]
        assert instance.deployment_name == "sandbox"  # pyright: ignore[reportAttributeAccessIssue]

    with gen_agent_instance(
        url=dagster_cloud_url,
        token="token",
        dagster_cloud_api_config={"deployments": ["sandbox", "prod"]},
    ) as instance:
        assert instance.deployment_names == ["sandbox", "prod"]  # pyright: ignore[reportAttributeAccessIssue]
        with pytest.raises(Exception):
            instance.deployment_name  # noqa: B018  # pyright: ignore[reportAttributeAccessIssue]

    with pytest.raises(Exception, match="Cannot set both deployment and deployments"):
        with gen_agent_instance(
            url=dagster_cloud_url,
            token="token",
            deployment="sandbox",
            dagster_cloud_api_config={"deployments": ["sandbox", "prod"]},
        ) as instance:
            pass


def test_instance_proxies(dagster_cloud_url):
    proxies = {
        "http": "http://proxy.example.com:8080",
        "https": "http://secureproxy.example.com:8090",
    }

    with gen_agent_instance(
        url=dagster_cloud_url,
        token="token",
        deployment="sandbox",
        dagster_cloud_api_config={"proxies": proxies},
    ) as instance:
        assert instance.graphql_client._proxies == proxies  # noqa: SLF001  # pyright: ignore[reportAttributeAccessIssue]


def test_instance_socket_options(dagster_cloud_url):
    with gen_agent_instance(
        url=dagster_cloud_url,
        token="token",
        deployment="sandbox",
    ) as instance:
        for session in [
            instance.client_managed_retries_requests_session,  # pyright: ignore[reportAttributeAccessIssue]
            instance.requests_managed_retries_session,  # pyright: ignore[reportAttributeAccessIssue]
        ]:
            adapter = session.get_adapter("https://")
            assert "socket_options" not in adapter.poolmanager.connection_pool_kw
        assert (
            instance.requests_managed_retries_session.get_adapter("https://").max_retries.total  # pyright: ignore[reportAttributeAccessIssue]
            == DEFAULT_RETRIES
        )

        assert (
            instance.client_managed_retries_requests_session.get_adapter(  # pyright: ignore[reportAttributeAccessIssue]
                "https://"
            ).max_retries.total
            == 0
        )
    with gen_agent_instance(
        url=dagster_cloud_url,
        token="token",
        deployment="sandbox",
        dagster_cloud_api_config={"socket_options": []},
    ) as instance:
        for session in [
            instance.client_managed_retries_requests_session,  # pyright: ignore[reportAttributeAccessIssue]
            instance.requests_managed_retries_session,  # pyright: ignore[reportAttributeAccessIssue]
        ]:
            adapter = session.get_adapter("https://")
            assert adapter.poolmanager.connection_pool_kw["socket_options"] == []

        assert (
            instance.requests_managed_retries_session.get_adapter("https://").max_retries.total  # pyright: ignore[reportAttributeAccessIssue]
            == DEFAULT_RETRIES
        )

        assert (
            instance.client_managed_retries_requests_session.get_adapter(  # pyright: ignore[reportAttributeAccessIssue]
                "https://"
            ).max_retries.total
            == 0
        )
    socket_options = [
        [6, 1, 1],
        ["SOL_SOCKET", "SO_KEEPALIVE", 1],
        ["IPPROTO_TCP", "TCP_KEEPCNT", 11],
    ]

    with gen_agent_instance(
        url=dagster_cloud_url,
        token="token",
        deployment="sandbox",
        dagster_cloud_api_config={"socket_options": socket_options},
    ) as instance:
        for session in [
            instance.client_managed_retries_requests_session,  # pyright: ignore[reportAttributeAccessIssue]
            instance.requests_managed_retries_session,  # pyright: ignore[reportAttributeAccessIssue]
        ]:
            adapter = session.get_adapter("https://")
            assert adapter.poolmanager.connection_pool_kw["socket_options"] == [
                (6, 1, 1),
                (socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1),
                (socket.IPPROTO_TCP, socket.TCP_KEEPCNT, 11),
            ]

        assert (
            instance.requests_managed_retries_session.get_adapter("https://").max_retries.total  # pyright: ignore[reportAttributeAccessIssue]
            == DEFAULT_RETRIES
        )

        assert (
            instance.client_managed_retries_requests_session.get_adapter(  # pyright: ignore[reportAttributeAccessIssue]
                "https://"
            ).max_retries.total
            == 0
        )


def test_instance_upload_snapshots_on_startup(dagster_cloud_url):
    with gen_agent_instance(
        url=dagster_cloud_url,
        token="token",
        deployment="sandbox",
    ) as instance:
        assert instance.user_code_launcher.upload_outdated_snapshots_on_startup  # pyright: ignore[reportAttributeAccessIssue]

    with gen_agent_instance(
        url=dagster_cloud_url,
        token="token",
        user_code_launcher_config={"upload_snapshots_on_startup": False},
    ) as instance:
        assert not instance.user_code_launcher.upload_outdated_snapshots_on_startup  # pyright: ignore[reportAttributeAccessIssue]


def test_instance_backcompat(dagster_cloud_url):
    with mock.patch("kubernetes.config.load_incluster_config"):
        with gen_agent_instance(
            url=dagster_cloud_url,
            token="token",
            deployment="sandbox",
            additional_config={
                "agent_queues": {"include_default_queue": False},
                "user_code_launcher": {
                    "module": "dagster_cloud.workspace.kubernetes",
                    "class": "K8sUserCodeLauncher",
                    "config": {
                        "dagster_home": "MY_DAGSTER_HOME",
                        "instance_config_map": "MY_INSTANCE_CONFIG_MAP",
                        "service_account_name": "MY_SERVICE_ACCOUNT_NAME",
                        "agent_metrics": {"enabled": True},
                    },
                },
            },
        ) as instance:
            assert instance.user_code_launcher.upload_outdated_snapshots_on_startup  # pyright: ignore[reportAttributeAccessIssue]
            assert not instance.agent_queues_config.include_default_queue  # pyright: ignore[reportAttributeAccessIssue]

            assert (
                "agent_queues"
                in instance.ref_for_deployment("prod").custom_instance_class_data.config_dict  # pyright: ignore[reportAttributeAccessIssue]
            )

            assert instance.user_code_launcher.agent_metrics_enabled  # pyright: ignore[reportAttributeAccessIssue]

            user_code_instance_ref = get_instance_ref_for_user_code(
                instance.ref_for_deployment("prod")  # pyright: ignore[reportAttributeAccessIssue]
            )

            assert (
                "agent_queues" not in user_code_instance_ref.custom_instance_class_data.config_dict  # pyright: ignore[reportOptionalMemberAccess]
            )
            assert (
                "agent_metrics"
                not in user_code_instance_ref.custom_instance_class_data.config_dict[  # pyright: ignore[reportOptionalMemberAccess]
                    "user_code_launcher"
                ]["config"]
            )

            user_code_instance = DagsterInstance.from_ref(user_code_instance_ref)
            assert not user_code_instance.user_code_launcher.agent_metrics_enabled  # pyright: ignore[reportAttributeAccessIssue]

            assert (
                user_code_instance.dagster_cloud_url  # pyright: ignore[reportAttributeAccessIssue]
                == DagsterInstance.from_ref(instance.ref_for_deployment("prod")).dagster_cloud_url  # pyright: ignore[reportAttributeAccessIssue]
            )



import logging
import os
import signal
import subprocess

from dagster._serdes import deserialize_value
from dagster._utils.interrupts import setup_interrupt_handlers
from dagster_cloud_cli.core.workspace import PexMetadata
from typer import Option, Typer

from dagster_cloud.pex.grpc.server.registry import PexS3Registry
from dagster_cloud.pex.grpc.server.server import run_multipex_server

app = Typer(hidden=True)


@app.command(short_help="Run multi-pex server to spin up subprocesses via PEX files")
def grpc(
    host: str = Option(default="localhost"),
    port: int | None = Option(default=None),
    socket: str | None = Option(default=None),
    max_workers: int | None = Option(default=None),
    local_pex_files_dir: str | None = Option(
        default="/tmp/pex-files", envvar="LOCAL_PEX_FILES_DIR"
    ),
    watchdog_run_interval: int | None = Option(default=30, envvar="WATCHDOG_RUN_INTERVAL"),
    enable_metrics: bool = Option(default=False, envvar="PEX_ENABLE_METRICS"),
):
    logger = logging.getLogger("dagster.multipex_server")

    run_multipex_server(
        port=port,
        socket=socket,
        host=host,
        logger=logger,
        max_workers=max_workers,
        local_pex_files_dir=local_pex_files_dir,
        watchdog_run_interval=watchdog_run_interval,
        enable_metrics=enable_metrics,
    )


@app.command(short_help="Execute a run with a PEX file")
def execute_run(
    input_json: str,
    pex_metadata_json: str,
    local_pex_files_dir: str | None = Option(
        default="/tmp/pex-files", envvar="LOCAL_PEX_FILES_DIR"
    ),
):
    setup_interrupt_handlers()
    pex_metadata = deserialize_value(pex_metadata_json, PexMetadata)
    executable = PexS3Registry(local_pex_files_dir).get_pex_executable(pex_metadata)

    run_process = subprocess.Popen(
        [
            executable.source_path,
            "-m",
            "dagster",
            "api",
            "execute_run",
            input_json,
        ],
        env={**os.environ.copy(), **executable.environ},
        cwd=executable.working_directory,
    )

    logger = logging.getLogger("dagster.pex_run")

    try:
        return_code = run_process.wait()
        if return_code != 0:
            raise Exception(f"PEX subprocess returned with exit code {return_code}")
    except KeyboardInterrupt:
        logger.info("Forwarding interrupt to PEX subprocess")
        run_process.send_signal(signal.SIGINT)
        run_process.wait()
        raise



from contextlib import contextmanager

from dagster._core.test_utils import instance_for_test
from dagster._utils.merger import merge_dicts


@contextmanager
def gen_agent_instance(
    url=None,
    token=None,
    timeout=None,
    deployment=None,
    python_logs=None,
    user_code_launcher_config=None,
    dagster_cloud_api_config=None,
    additional_config=None,
):
    with instance_for_test(
        merge_dicts(
            {
                "instance_class": {
                    "module": "dagster_cloud",
                    "class": "DagsterCloudAgentInstance",
                },
                "user_code_launcher": {
                    "module": "dagster_cloud.workspace.user_code_launcher",
                    "class": "ProcessUserCodeLauncher",
                    "config": user_code_launcher_config or {},
                },
                "dagster_cloud_api": merge_dicts(
                    {"url": url} if url else {},
                    {"agent_token": token} if token else {},
                    ({"timeout": timeout} if timeout else {}),
                    {"deployment": deployment} if deployment else {},
                    dagster_cloud_api_config or {},
                ),
            },
            {"python_logs": python_logs} if python_logs else {},
            (additional_config or {}),
        )
    ) as instance:
        yield instance



from dagster._core.storage.noop_compute_log_manager import NoOpComputeLogManager
from dagster._core.test_utils import instance_for_test
from dagster_cloud.storage.compute_logs import CloudComputeLogManager


def test_cloud_compute_log_manager_configured():
    with instance_for_test(
        {
            "instance_class": {
                "module": "dagster_cloud",
                "class": "DagsterCloudAgentInstance",
            },
            "user_code_launcher": {
                "module": "dagster_cloud.workspace.user_code_launcher",
                "class": "ProcessUserCodeLauncher",
            },
            "dagster_cloud_api": {
                "url": "https://agent-api-test.dagster.cloud",
                "agent_token": "token",
            },
        }
    ) as instance:
        assert isinstance(instance.compute_log_manager, CloudComputeLogManager)


def test_cloud_compute_log_manager_configured_disabled():
    with instance_for_test(
        {
            "instance_class": {
                "module": "dagster_cloud",
                "class": "DagsterCloudAgentInstance",
            },
            "user_code_launcher": {
                "module": "dagster_cloud.workspace.user_code_launcher",
                "class": "ProcessUserCodeLauncher",
            },
            "dagster_cloud_api": {
                "url": "https://agent-api-test.dagster.cloud",
                "agent_token": "token",
            },
            "compute_logs": {
                "module": "dagster.core.storage.noop_compute_log_manager",
                "class": "NoOpComputeLogManager",
            },
        }
    ) as instance:
        assert isinstance(instance.compute_log_manager, NoOpComputeLogManager)


def test_cloud_compute_log_manager_streaming():
    with instance_for_test(
        {
            "instance_class": {
                "module": "dagster_cloud",
                "class": "DagsterCloudAgentInstance",
            },
            "user_code_launcher": {
                "module": "dagster_cloud.workspace.user_code_launcher",
                "class": "ProcessUserCodeLauncher",
            },
            "dagster_cloud_api": {
                "url": "https://agent-api-test.dagster.cloud",
                "agent_token": "token",
            },
            "compute_logs": {
                "module": "dagster_cloud",
                "class": "CloudComputeLogManager",
                "config": {
                    "upload_interval": 10,
                },
            },
        }
    ) as instance:
        assert isinstance(instance.compute_log_manager, CloudComputeLogManager)
        assert instance.compute_log_manager.upload_interval == 10



import base64
import json
import time
from contextlib import contextmanager

# A minimal copy of the standard OTEL API


class Span:
    def __init__(self, name: str):
        self._name = name
        self._attrs = {}
        self._start_time = time.time()
        self._end_time = None

    def set_attribute(self, key: str, value: int | str):
        self._attrs[key] = value

    def end(self):
        self._end_time = time.time()

    def get_serialized(self) -> str:
        # base64 to make it safe for a http header (no newlines, few special chars)
        return base64.b64encode(
            json.dumps(
                {
                    "_name": self._name,
                    "_duration": self._end_time - self._start_time
                    if self._end_time is not None
                    else None,
                    **self._attrs,
                },
                separators=(",", ":"),
                indent=None,
            ).encode("utf-8")
        ).decode()


class Tracer:
    def __init__(self):
        self._spans: list[Span] = []

    @contextmanager
    def start_span(self, name: str):
        span = Span(name)
        try:
            yield span
        finally:
            span.end()
            self._spans.append(span)

    def pop_serialized_span(self) -> str | None:
        try:
            span = self._spans.pop(0)
        except IndexError:
            return None
        else:
            return span.get_serialized()



import base64
import json
import platform
import re
import time
from contextlib import nullcontext
from email.utils import formatdate

import freezegun
import pytest
import responses
from dagster._core.events import DagsterEvent, DagsterEventType, EngineEventData
from dagster._core.events.log import EventLogEntry
from dagster._core.test_utils import environ
from dagster_cloud.storage.client import DEFAULT_RETRIES
from dagster_cloud.version import __version__
from dagster_cloud_cli.core.errors import DagsterCloudAgentServerError
from dagster_cloud_cli.core.headers.impl import (
    API_TOKEN_HEADER,
    DAGSTER_CLOUD_VERSION_HEADER,
    PYTHON_VERSION_HEADER,
)
from requests.exceptions import ConnectionError as RequestsConnectionError

from dagster_cloud_tests import gen_agent_instance


@pytest.fixture(
    params=["graphql", "http"],
)
def api_protocol(request):
    with (
        environ({"DAGSTER_CLOUD_STORE_EVENT_OVER_HTTP": "1"})
        if request.param == "http"
        else nullcontext()
    ):
        yield request.param


@pytest.fixture
def graphql_url(dagster_cloud_url, api_protocol):
    if api_protocol == "graphql":
        return dagster_cloud_url + "/graphql"
    else:
        return dagster_cloud_url + "/store_events"


@pytest.fixture
def agent_instance(agent_token, dagster_cloud_url):
    with gen_agent_instance(dagster_cloud_url, agent_token) as instance:
        yield instance


@pytest.fixture
def proxy_graphql_client(agent_instance):
    return agent_instance.graphql_client


def test_graphql_client_headers(proxy_graphql_client, agent_token):
    headers = proxy_graphql_client.headers
    assert headers[DAGSTER_CLOUD_VERSION_HEADER] == __version__
    assert headers[PYTHON_VERSION_HEADER] == platform.python_version()
    assert headers[API_TOKEN_HEADER] == agent_token


@pytest.fixture
def fake_sleeps(monkeypatch):
    sleeps = []

    def fake_sleep(s):
        sleeps.append(s)

    monkeypatch.setattr(time, "sleep", fake_sleep)

    yield sleeps


def create_test_event_log_record(message: str, run_id):
    return


def _store_event(agent_instance):
    event_log_entry = EventLogEntry(
        error_info=None,
        user_message="hello",
        level="debug",
        run_id="fake-run-id",
        timestamp=time.time(),
        dagster_event=DagsterEvent(
            DagsterEventType.ENGINE_EVENT.value,
            "nonce",
            event_specific_data=EngineEventData.in_process(999),
        ),
    )
    agent_instance.event_log_storage.store_event(event_log_entry)


@responses.activate
def test_graphql_client_backoff_succeeds(agent_instance, fake_sleeps, graphql_url):
    num_callbacks = {"num": 0}
    good_result = {"data": "youdidit"}

    def my_callback(request):
        num_callbacks["num"] = num_callbacks["num"] + 1
        if num_callbacks["num"] > 3:
            return (200, {}, json.dumps(good_result))

        return (502, {}, "")

    responses.add_callback(responses.POST, graphql_url, callback=my_callback)

    _store_event(agent_instance)

    assert num_callbacks["num"] == 4
    assert fake_sleeps == [1, 2, 4]


@responses.activate
def test_graphql_client_429(agent_instance, fake_sleeps, graphql_url):
    num_callbacks = {"num": 0}
    good_result = {"data": "youdidit"}

    def my_callback(request):
        num_callbacks["num"] = num_callbacks["num"] + 1
        if num_callbacks["num"] > 3:
            return (200, {}, json.dumps(good_result))

        return (429, {"Retry-After": "10"}, "")

    responses.add_callback(responses.POST, graphql_url, callback=my_callback)

    _store_event(agent_instance)

    assert num_callbacks["num"] == 4
    assert fake_sleeps == [10, 10, 10]


@responses.activate
def test_graphql_client_429_date(agent_instance, fake_sleeps, graphql_url):
    num_callbacks = {"num": 0}
    good_result = {"data": "youdidit"}

    with freezegun.freeze_time("2023-01-01"):
        now = time.time()
        future = now + 10

        def my_callback(request):
            num_callbacks["num"] = num_callbacks["num"] + 1
            if num_callbacks["num"] > 1:
                return (200, {}, json.dumps(good_result))

            return (429, {"Retry-After": formatdate(future, usegmt=True)}, "")

        responses.add_callback(responses.POST, graphql_url, callback=my_callback)

        _store_event(agent_instance)

        assert num_callbacks["num"] == 2
        assert fake_sleeps == [10]


@responses.activate
def test_graphql_client_429_bad_retry_after(agent_instance, fake_sleeps, graphql_url):
    num_callbacks = {"num": 0}
    good_result = {"data": "youdidit"}

    def my_callback(request):
        num_callbacks["num"] = num_callbacks["num"] + 1
        if num_callbacks["num"] > 3:
            return (200, {}, json.dumps(good_result))
        elif num_callbacks["num"] > 2:
            return (429, {"Retry-After": "junk"}, "")
        elif num_callbacks["num"] > 1:
            return (429, {"Retry-After": "-1"}, "")

        return (429, {}, "")

    responses.add_callback(responses.POST, graphql_url, callback=my_callback)

    _store_event(agent_instance)

    assert num_callbacks["num"] == 4
    assert fake_sleeps == [1, 2, 4]


@responses.activate
def test_graphql_client_backoff_fails(agent_instance, fake_sleeps, graphql_url):
    num_callbacks = {"num": 0}

    def my_callback(request):
        num_callbacks["num"] = num_callbacks["num"] + 1
        return (502, {}, "")

    responses.add_callback(responses.POST, graphql_url, callback=my_callback)

    with pytest.raises(
        DagsterCloudAgentServerError, match=re.escape("too many 502 error responses")
    ):
        _store_event(agent_instance)

    assert num_callbacks["num"] == DEFAULT_RETRIES + 1
    assert fake_sleeps == [1, 2, 4, 8, 16, 32]


@responses.activate
@pytest.mark.parametrize(
    "error_msg",
    [
        "SSLError",
        "ConnectTimeoutError",
        "Connection reset by peer",
    ],
)
def test_graphql_client_connection_retry_succeeds(
    agent_instance, fake_sleeps, error_msg, graphql_url
):
    num_callbacks = {"num": 0}
    good_result = {"data": "youdidit"}

    def my_callback(request):
        num_callbacks["num"] = num_callbacks["num"] + 1
        if num_callbacks["num"] > 5:
            return (200, {}, json.dumps(good_result))
        raise RequestsConnectionError(error_msg)

    responses.add_callback(responses.POST, graphql_url, callback=my_callback)

    _store_event(agent_instance)

    assert num_callbacks["num"] == 6
    assert fake_sleeps == [1, 2, 4, 8, 16]


@responses.activate
def test_graphql_client_connection_reset_retry_fails(agent_instance, fake_sleeps, graphql_url):
    num_callbacks = {"num": 0}

    def my_callback(request):
        num_callbacks["num"] = num_callbacks["num"] + 1
        raise RequestsConnectionError("Connection reset by peer")

    responses.add_callback(responses.POST, graphql_url, callback=my_callback)

    with pytest.raises(DagsterCloudAgentServerError, match=re.escape("Connection reset by peer")):
        _store_event(agent_instance)

    assert num_callbacks["num"] == DEFAULT_RETRIES + 1
    assert fake_sleeps == [1, 2, 4, 8, 16, 32]


@responses.activate
def test_graphql_client_metrics(agent_instance, fake_sleeps, graphql_url, monkeypatch):
    metric_headers = []

    def my_callback(request):
        metric_header = request.headers.get("Dagster-Cloud-Metric")
        if metric_header:
            metric_headers.append(json.loads(base64.b64decode(metric_header)))
        else:
            metric_headers.append(None)

        return (200, {}, json.dumps({"data": "good"}))

    responses.add_callback(responses.POST, graphql_url, callback=my_callback)
    monkeypatch.setenv("DISABLE_DAGSTER_CLOUD_STORE_EVENT_SEND_METRICS", "1")
    _store_event(agent_instance)
    monkeypatch.delenv("DISABLE_DAGSTER_CLOUD_STORE_EVENT_SEND_METRICS")
    _store_event(agent_instance)
    _store_event(agent_instance)
    monkeypatch.setenv("DISABLE_DAGSTER_CLOUD_STORE_EVENT_SEND_METRICS", "1")
    _store_event(agent_instance)

    assert metric_headers[0] is None
    assert metric_headers[1] is None, (
        "First event after we start recording will not have metric of previous event"
    )
    assert metric_headers[2], "expected metric of previous event"
    assert metric_headers[2]["_name"] == "store-event"
    assert metric_headers[2]["_duration"] > 0
    assert metric_headers[2]["event_type"] == "ENGINE_EVENT"
    assert metric_headers[2]["run_id"] == "fake-run-id"
    assert metric_headers[3] is None, "No metric sent after env turned off"



from datetime import timedelta
from typing import TYPE_CHECKING, Any, cast

import pytest
from dagster import Definitions, RunRequest, build_schedule_context
from dagster._core.test_utils import freeze_time
from dagster._time import create_datetime, get_current_datetime
from dagster_cloud.dagster_insights import create_snowflake_insights_asset_and_schedule

if TYPE_CHECKING:
    from collections.abc import Iterable


@pytest.mark.parametrize("additional_schedule_offset", [0, 1, 2])
def test_schedule_definition(additional_schedule_offset: int) -> Any:
    for hour in range(0, 24):
        for minute in range(0, 60, 5):
            freeze_datetime = create_datetime(2023, 2, 27, hour, minute, 0, 0, tz="UTC")

            with freeze_time(freeze_datetime):
                start_date = (get_current_datetime() - timedelta(days=5)).replace(
                    minute=0, second=0, microsecond=0
                )

                asset_and_schedule = create_snowflake_insights_asset_and_schedule(
                    start_date=start_date,
                    name="insights_test",
                    snowflake_resource_key="snowflake_insights",
                    partition_end_offset_hrs=additional_schedule_offset,
                )

                defs = Definitions(
                    assets=asset_and_schedule.assets,
                    schedules=[asset_and_schedule.schedule],
                    resources={"snowflake_insights": None},
                )
                resolved_schedule = defs.resolve_schedule_def("snowflake_insights_import_schedule")

                execution_time = (get_current_datetime() - timedelta(hours=1)).replace(
                    minute=59, second=0, microsecond=0
                )
                result = resolved_schedule(
                    build_schedule_context(scheduled_execution_time=execution_time)
                )
                assert isinstance(result, RunRequest)

                assert result.partition_key == (
                    execution_time - timedelta(hours=1 + additional_schedule_offset)
                ).strftime("%Y-%m-%d-%H:00")


@pytest.mark.parametrize("additional_schedule_offset", [0, 1, 2])
def test_schedule_definition_group(additional_schedule_offset: int) -> Any:
    for group_hours in [2, 4, 8, 12, 24]:
        for hour in range(0, 24):
            for minute in range(0, 60, 5):
                freeze_datetime = create_datetime(2023, 2, 27, hour, minute, 0, 0, tz="UTC")

                with freeze_time(freeze_datetime):
                    start_date = (get_current_datetime() - timedelta(days=5)).replace(
                        minute=0, second=0, microsecond=0
                    )

                    asset_and_schedule = create_snowflake_insights_asset_and_schedule(
                        start_date=start_date,
                        name="insights_test",
                        snowflake_resource_key="snowflake_insights",
                        schedule_batch_size_hrs=group_hours,
                        partition_end_offset_hrs=additional_schedule_offset,
                    )

                    defs = Definitions(
                        assets=asset_and_schedule.assets,
                        schedules=[asset_and_schedule.schedule],
                        resources={"snowflake_insights": None},
                    )
                    resolved_schedule = defs.resolve_schedule_def(
                        f"snowflake_insights_import_schedule_{group_hours}_hrs"
                    )

                    execution_time = (get_current_datetime() - timedelta(hours=1)).replace(
                        minute=59, second=0, microsecond=0
                    )
                    raw_results = resolved_schedule(
                        build_schedule_context(scheduled_execution_time=execution_time)
                    )
                    assert raw_results is not None
                    results = list(cast("Iterable[RunRequest]", raw_results))

                    assert len(results) == 1
                    assert isinstance(results[0], RunRequest)

                    # Ensure we request a single run for the entire group
                    assert results[0].partition_key_range and results[
                        0
                    ].partition_key_range.start == (
                        execution_time
                        - timedelta(hours=1 + additional_schedule_offset + group_hours - 1)
                    ).strftime("%Y-%m-%d-%H:00")

                    assert results[0].partition_key_range.end == (
                        execution_time - timedelta(hours=1 + additional_schedule_offset)
                    ).strftime("%Y-%m-%d-%H:00")



import uuid

from dagster._core.errors import DagsterInvariantViolationError


def get_hardcoded_test_agent_token(organization_name) -> str:
    return f"agent:{organization_name}:hardcoded"


def get_hardcoded_test_user_token(organization_name, user_name) -> str:
    return f"user:{organization_name}:{user_name}"


def get_organization_public_id_from_api_token(api_token: str) -> str | None:
    split_token = api_token.split(":")
    if len(split_token) != 4:
        raise Exception("Could not derive organization from api token")

    return split_token[2]


def decode_region_from_uuid(regional_token: str) -> str | None:
    try:
        regional_uuid = uuid.UUID(regional_token)
    except ValueError:
        # if it's not an actual uuid, we can't decode region
        return None

    # custom uuids contain region subdomains in the first 2 bytes
    if regional_uuid.version != 8 or regional_uuid.variant != uuid.RFC_4122:
        return None

    uuid_bytes = regional_uuid.bytes
    return uuid_bytes[:2].decode("ascii")


def decode_agent_token(agent_token: str) -> tuple[str | None, str | None]:
    split_token = agent_token.split(":")

    # Legacy agent token format - organization must be specified in dagster.yaml
    if len(split_token) == 1:
        return None, None

    token_type, *token = split_token

    if token_type == "user":
        raise DagsterInvariantViolationError(
            "Agent was configured with a user token, but agents can only authenticate with "
            "Dagster Cloud when configured with an agent token. "
            "Generate a new agent token in Dagster Cloud."
        )

    # token format: agent:<org>:<uuid>
    organization, uuid_str = token
    return organization, decode_region_from_uuid(uuid_str)



from unittest import mock

from dagster_cloud.api.dagster_cloud_api import (
    DagsterCloudUploadLocationData,
    DagsterCloudUploadWorkspaceEntry,
)
from dagster_cloud.instance import DagsterCloudAgentInstance
from dagster_cloud_backend.analytics.events import AnalyticsEvent
from dagster_cloud_cli.core.workspace import CodeLocationDeployData


def test_add_location_adds_segment_event(
    agent_instance_local_ursula: DagsterCloudAgentInstance,
    host_instance,
    segment_analytics,  # (fixture)
) -> None:
    # Verify that the agent reconciling and uploading a new location triggers Segment

    agent_instance = agent_instance_local_ursula

    user_code_launcher = agent_instance.user_code_launcher

    code_location_deploy_data = CodeLocationDeployData(
        python_file=__file__, image="dagster/test_image:bar"
    )

    host_instance.cloud_storage.add_location(
        "foo", code_location_deploy_data=code_location_deploy_data
    )

    user_code_launcher._update_workspace_entry(  # noqa: SLF001
        "sandbox",
        DagsterCloudUploadWorkspaceEntry(
            location_name="foo",
            code_location_deploy_data=code_location_deploy_data,
            upload_location_data=DagsterCloudUploadLocationData(
                upload_repository_datas=[],
                container_image=None,
                executable_path=None,
            ),
            serialized_error_info=None,
        ),
        server_or_error=mock.MagicMock(),
    )

    assert len(segment_analytics.track_calls) == 1
    assert segment_analytics.track_calls[0]["user_id"] != "agent"
    assert segment_analytics.track_calls[0]["event"] == AnalyticsEvent.CODE_LOCATION_ADD
    assert (
        segment_analytics.track_calls[0]["properties"]["groupId"] == host_instance.organization_id
    )



from setuptools import find_packages, setup

setup(
    name="repo-1",
    packages=find_packages(exclude=["repo_1_tests"]),
    install_requires=[
        "dagster",
        "dagster-cloud",
        "flask",  # provides the flask binary
    ],
    extras_require={"dev": ["dagster-webserver", "pytest"]},
)



import pytest
from dagster._core.remote_origin import (
    RegisteredCodeLocationOrigin,
    RemoteJobOrigin,
    RemoteRepositoryOrigin,
)
from dagster_cloud.workspace.ecs.utils import (
    get_run_task_definition_family,
    get_server_task_definition_family,
)


@pytest.fixture
def job_origin():
    return RemoteJobOrigin(
        repository_origin=RemoteRepositoryOrigin(
            repository_name="the_repo",
            code_location_origin=RegisteredCodeLocationOrigin(location_name="the_location"),
        ),
        job_name="the_job",
    )


@pytest.fixture
def long_job_origin():
    long_job_name = "a" * 512
    long_repo_name = "b" * 512
    long_location_name = "c" * 512

    return RemoteJobOrigin(
        repository_origin=RemoteRepositoryOrigin(
            repository_name=long_repo_name,
            code_location_origin=RegisteredCodeLocationOrigin(location_name=long_location_name),
        ),
        job_name=long_job_name,
    )


@pytest.fixture
def other_long_job_origin():
    long_job_name = "d" * 512
    long_repo_name = "e" * 512
    long_location_name = "f" * 512

    return RemoteJobOrigin(
        repository_origin=RemoteRepositoryOrigin(
            repository_name=long_repo_name,
            code_location_origin=RegisteredCodeLocationOrigin(location_name=long_location_name),
        ),
        job_name=long_job_name,
    )


def test_task_definition_family_long_location(long_job_origin, other_long_job_origin):
    family = get_run_task_definition_family("foo", "hooli", "prod", long_job_origin)

    assert (
        family
        == "foo_hooli_prod_ccccccccccccccccccccccc_d9023790_bbbbbbbbbbbbbbbbbbbbbbb_3956139d_aaaaaaaaaaaaaaaaaaaaaaa_164557fa"
    )

    assert family != get_run_task_definition_family("foo", "hooli", "prod", other_long_job_origin)


def test_org_deployment_scope(job_origin):
    assert (
        get_run_task_definition_family("foo", "hooli", "prod", job_origin)
        == "foo_hooli_prod_the_location_66c360f2_the_repo_b9c5532e_the_job_38cc9a96"
    )

    # different deployment = new family
    assert get_run_task_definition_family(
        "foo", "hooli", "prod", job_origin
    ) != get_run_task_definition_family("foo", "hooli", "dev", job_origin)

    # different org = new family
    assert get_run_task_definition_family(
        "foo", "hooli", "prod", job_origin
    ) != get_run_task_definition_family("foo", "acme", "prod", job_origin)

    # different prefix = new family

    assert get_run_task_definition_family(
        "foo", "hooli", "prod", job_origin
    ) != get_run_task_definition_family("bar", "hooli", "prod", job_origin)

    # same everything = same
    assert get_run_task_definition_family(
        "foo", "hooli", "prod", job_origin
    ) == get_run_task_definition_family("foo", "hooli", "prod", job_origin)


def test_server_task_definition_family_long_location():
    long_location_name = "abcd" * 255
    different_long_location_name = "abcd" * 512

    assert get_server_task_definition_family(
        "foo", "hooli", "prod", long_location_name
    ) != get_server_task_definition_family("foo", "hooli", "prod", different_long_location_name)


def test_server_task_definition_family_short_location():
    short_location_name = "abcd"
    other_short_location_name = "efgh"

    assert get_server_task_definition_family(
        "foo", "hooli", "prod", short_location_name
    ) != get_server_task_definition_family("foo", "hooli", "prod", other_short_location_name)


def test_server_org_deployment_scope():
    location_name = "abcd"

    # different deployment = new family
    assert get_server_task_definition_family(
        "foo", "hooli", "prod", location_name
    ) != get_server_task_definition_family("foo", "hooli", "dev", location_name)

    # different org = new family
    assert get_server_task_definition_family(
        "foo", "hooli", "prod", location_name
    ) != get_server_task_definition_family("foo", "acme", "prod", location_name)

    # same everything = same
    assert get_server_task_definition_family(
        "foo", "hooli", "prod", location_name
    ) == get_server_task_definition_family("foo", "hooli", "prod", location_name)



import pytest


@pytest.fixture(autouse=True)
def env(monkeypatch):
    # Prevent boto clients from actually connecting to AWS;
    # real connections should only happen in integration tests
    monkeypatch.setenv("AWS_REGION", "us-east-1")
    monkeypatch.setenv("AWS_DEFAULT_REGION", "us-east-1")
    monkeypatch.setenv("AWS_ACCESS_KEY_ID", "test")
    monkeypatch.setenv("AWS_SECRET_ACCESS_KEY", "test")



import subprocess
import tempfile
from pathlib import Path

import dagster_cloud_cli
import pytest
from dagster_cloud_cli.entrypoint import app
from typer.testing import CliRunner

runner = CliRunner()


@pytest.fixture(autouse=True)
def mock_gql(mocker):
    yield mocker.patch("dagster_cloud_cli.commands.serverless.metrics.gql", autospec=True)


def build_dagster_project(code_directory):
    with tempfile.TemporaryDirectory() as build_dir:
        result = runner.invoke(
            app,
            [
                "serverless",
                "build-python-executable",
                "--python-version=3.12",
                str(code_directory),
                build_dir,
            ],
            env={
                "DAGSTER_CLOUD_API_TOKEN": "fake",
                "DAGSTER_CLOUD_ORGANIZATION": "fake",
                "DAGSTER_CLOUD_DEPLOYMENT": "fake",
            },
        )
        assert result.exit_code == 0, f"Failed to build {code_directory}: {result.output}"
        assert "Built" in result.stdout

        build_path = Path(build_dir)
        pex_files = [
            path for path in build_path.iterdir() if path.name.endswith(".pex") and path.is_file()
        ]
        assert len(pex_files) == 2


def test_build_python_executable(sample_repos_dir):
    build_dagster_project(str(sample_repos_dir / "repo-1"))


def test_build_quickstart_etl():
    import dagster

    build_dagster_project(Path(dagster.__file__).parents[3] / "examples/quickstart_etl")


def test_build_methods(sample_repos_dir, mocker):
    # mock pex builder function to allow forcing failure using deps_build_stderr
    orig_build_pex = dagster_cloud_cli.core.pex_builder.util.build_pex  # pyright: ignore[reportAttributeAccessIssue]
    deps_build_stderr = None

    def build_pex_fail_deps(
        sources_directories: list[str],
        requirements_filepaths: list[str],
        pex_flags: list[str],
        output_pex_path: str,
        pex_root: str | None = None,
    ):
        if requirements_filepaths and deps_build_stderr is not None:
            return subprocess.CompletedProcess(["testmock"], 1, b"stdout", deps_build_stderr)
        return orig_build_pex(
            sources_directories, requirements_filepaths, pex_flags, output_pex_path, pex_root
        )

    build_pex = mocker.patch(
        "dagster_cloud_cli.core.pex_builder.util.build_pex", side_effect=build_pex_fail_deps
    )
    docker_runner_dagster_cloud = mocker.patch(
        "dagster_cloud_cli.core.docker_runner.run_dagster_cloud", autospec=True
    )

    def build_repo(extra_args=[]):
        code_directory = str(sample_repos_dir / "repo-1")
        with tempfile.TemporaryDirectory() as build_dir:
            return runner.invoke(
                app,
                [
                    "serverless",
                    "build-python-executable",
                    "--python-version=3.12",
                    str(code_directory),
                    build_dir,
                ]
                + extra_args,
                env={
                    "DAGSTER_CLOUD_API_TOKEN": "fake",
                    "DAGSTER_CLOUD_ORGANIZATION": "fake",
                    "DAGSTER_CLOUD_DEPLOYMENT": "fake",
                },
            )

    # build fails for unknown reason: should propagate
    deps_build_stderr = b"stderr123"
    result = build_repo(["--build-method=docker-fallback"])
    assert result.exit_code != 0, f"failure did not propgate: {result.output}"
    assert "stderr123" in result.output
    build_pex.assert_called_once()
    docker_runner_dagster_cloud.assert_not_called()

    # build fails due to ResolutionImpossible: should try and fallback to docker
    build_pex.reset_mock()
    docker_runner_dagster_cloud.reset_mock()
    deps_build_stderr = b"ResolutionImpossible"
    result = build_repo(["--build-method=docker-fallback"])
    build_pex.assert_called_once()
    docker_runner_dagster_cloud.assert_called_once()
    call_args = docker_runner_dagster_cloud.call_args
    assert "build-python-deps" in call_args.kwargs["run_args"]

    # when --build-method==local, don't fallback to docker
    build_pex.reset_mock()
    docker_runner_dagster_cloud.reset_mock()
    deps_build_stderr = b"ResolutionImpossible"
    result = build_repo(["--build-method=local"])
    build_pex.assert_called_once()
    docker_runner_dagster_cloud.assert_not_called()
    assert result.exit_code != 0, f"failure did not propgate: {result.output}"

    # when --build-method=local and no failure, dont fallback to docker
    build_pex.reset_mock()
    docker_runner_dagster_cloud.reset_mock()
    deps_build_stderr = None
    result = build_repo(["--build-method=local"])
    build_pex.assert_called()  # would be called twice - once for source and once for deps
    docker_runner_dagster_cloud.assert_not_called()
    assert result.exit_code == 0

    # when --build-method==docker, don't try to build locally, just in docker
    build_pex.reset_mock()
    docker_runner_dagster_cloud.reset_mock()
    deps_build_stderr = b"stderr456"
    result = build_repo(["--build-method=docker"])
    build_pex.assert_not_called()  # local build not attempted
    docker_runner_dagster_cloud.assert_called_once()



import dagster_cloud
import dagster_cloud_cli
from dagster._core.libraries import DagsterLibraryRegistry


def test_library_registry():
    library_registry = DagsterLibraryRegistry.get()
    assert library_registry["dagster-cloud"] == dagster_cloud.__version__
    assert library_registry["dagster-cloud-cli"] == dagster_cloud_cli.__version__



import re

# Number-only regex.
NUMBERS_ONLY_REGEX = re.compile(r"^\d+$")
# Fractional decimal-allowed number regex.
FRACTIONAL_REGEX = re.compile(r"^\d*\.?\d+$")
# K8s memory is in bytes when a plain number is provided, and in quantity strings when a string is provided.
K8S_MEM_QUANTITY_REGEX = re.compile(r"^(\d+)([EPTGMK])(i)?$")
# K8s CPU is in vCPUs when a number is provided, and in CPU units when a string ending in "m" is provided.
K8S_CPU_QUANTITY_REGEX = re.compile(r"^(\d+)(m?)$")
# ECS memory is MB when in integer, and GB when in string. Strings can contain spaces.
ECS_MEM_GB_REGEX = re.compile(r"^(\d*\.?\d+)(\s+)?GB$")
# ECS CPU is in CPU units when a number, and in vCPU when a string. Strings can contain spaces, and vcpu can be either lower or uppercase. We want to capture whether it's provided as a number or in terms of vCPUs.
ECS_CPU_VCPU_REGEX = re.compile(r"^(\d*\.?\d+)(\s+)?(v)?CPU$", flags=re.IGNORECASE)


def interpret_k8s_mem_str_as_bytes(mem_str: str | None) -> int | None:
    """Interpret a k8s memory quantity string as bytes."""
    # If the string is not provided, return None.
    if mem_str is None:
        return None

    # If the string is a number, then we interpret it as number of bytes.
    if NUMBERS_ONLY_REGEX.match(mem_str):
        return int(mem_str)

    # If the string is a quantity string, then we need to figure out the unit from the regex.
    match = K8S_MEM_QUANTITY_REGEX.match(mem_str)
    if match is None:
        raise Exception(f"Invalid k8s memory quantity string {mem_str}")

    # The first group is the number, the second group is the unit.
    num = int(match.group(1))
    unit = match.group(2)
    is_international_system = match.group(3) is not None

    if unit == "E":
        return num * 10**18 if not is_international_system else num * 2**60
    elif unit == "P":
        return num * 10**15 if not is_international_system else num * 2**50
    elif unit == "T":
        return num * 10**12 if not is_international_system else num * 2**40
    elif unit == "G":
        return num * 10**9 if not is_international_system else num * 2**30
    elif unit == "M":
        return num * 10**6 if not is_international_system else num * 2**20
    elif unit == "K":
        return num * 10**3 if not is_international_system else num * 2**10
    else:
        raise Exception(f"Invalid k8s memory quantity string {mem_str}")


def interpret_k8s_cpu_str_as_millicpus(cpu_str: str | None) -> int | None:
    """Interpret a k8s CPU quantity string as millicpus."""
    # If the string is not provided, return None.
    if cpu_str is None:
        return None

    # If the string is a number, then we interpret it as a fractional number of cpus.
    if FRACTIONAL_REGEX.match(cpu_str):
        return int(
            float(cpu_str) * 1000
        )  # K8s doesn't allow more precise than .001 CPU, meaning this will always be a whole number.

    # If the string is a quantity string, then we check to make sure it matches the millicpu syntax. Then we return the number.
    match = K8S_CPU_QUANTITY_REGEX.match(cpu_str)
    if match is None:
        raise Exception(f"Invalid k8s CPU quantity string {cpu_str}")

    # The first group is the number, the second group is the unit.
    return int(match.group(1))


def interpret_ecs_mem_str_as_bytes(mem_str: str | None) -> int | None:
    """Interpret an ECS memory string as bytes."""
    # If the string is not provided, return None.
    if mem_str is None:
        return None

    # If the string is a number, then we interpret it as number of MiB (mebibytes).
    if NUMBERS_ONLY_REGEX.match(mem_str):
        return int(mem_str) * 1024**2

    # If the string is a quantity string, then the unit should be GB.
    match = ECS_MEM_GB_REGEX.match(mem_str)
    if match is None:
        raise Exception(f"Invalid ECS memory string {mem_str}")

    # The first group is the number.
    num = int(match.group(1))
    return num * 1024**3


def interpret_ecs_cpu_str_as_millicpus(cpu_str: str | None) -> float | None:
    """Interpret an ECS CPU string as millicpus."""
    # If the string is not provided, return None.
    if cpu_str is None:
        return None

    # If the string is a number, then we interpret it as a whole number quantity of milliCPUs.
    if NUMBERS_ONLY_REGEX.match(cpu_str):
        return int(cpu_str)

    # If the string is a quantity string, then the unit should be vCPUs.
    match = ECS_CPU_VCPU_REGEX.match(cpu_str)
    if match is None:
        raise Exception(f"Invalid ECS CPU string {cpu_str}")

    # The first group is the number.
    num = float(match.group(1))
    return num * 1000



from pathlib import Path

import dagster as dg
from dagster.components.component.state_backed_component import StateBackedComponent
from dagster.components.utils.defs_state import DefsStateConfig
from dagster_shared.serdes.objects.models.defs_state_info import DefsStateManagementType


class SampleStateBackedComponent(StateBackedComponent, dg.Model, dg.Resolvable):
    @property
    def defs_state_config(self) -> DefsStateConfig:
        return DefsStateConfig(
            key=self.__class__.__name__,
            management_type=DefsStateManagementType.VERSIONED_STATE_STORAGE,
            refresh_if_dev=True,
        )

    def build_defs_from_state(
        self, context: dg.ComponentLoadContext, state_path: Path | None
    ) -> dg.Definitions:
        assert state_path is not None
        with open(state_path) as f:
            state = f.read()
        assert state == "hi"

        @dg.asset(name=state)
        def the_asset(): ...

        return dg.Definitions(assets=[the_asset])

    async def write_state_to_path(self, state_path: Path) -> None:
        state_path.write_text("hi")



import os
from pathlib import Path
from typing import cast

import pytest
from dagster import AssetsDefinition, load_assets_from_modules
from dagster._core.definitions.metadata import (
    LocalFileCodeReference,
    UrlCodeReference,
    with_source_code_references,
)
from dagster._core.definitions.metadata.source_code import FilePathMapping
from dagster._core.types.loadable_target_origin import (
    LoadableTargetOrigin,
    enter_loadable_target_origin_load_context,
)
from dagster._utils import file_relative_path
from dagster._utils.env import environ
from dagster_cloud.metadata.source_code import link_code_references_to_git_if_cloud

# path of the `dagster` package on the filesystem
DAGSTER_CLOUD_PACKAGE_PATH = os.path.normpath(file_relative_path(__file__, "../../"))
GIT_ROOT_PATH = os.path.normpath(os.path.join(DAGSTER_CLOUD_PACKAGE_PATH, "../../../"))

# path of the current file relative to the `dagster` package root
PATH_IN_PACKAGE = "/dagster_cloud_tests/metadata_tests/"

# {path to module}:{path to file relative to module root}:{line number}
EXPECTED_ORIGINS = {
    "james_brown": DAGSTER_CLOUD_PACKAGE_PATH + PATH_IN_PACKAGE + "asset_package/__init__.py:12",
    "chuck_berry": (
        DAGSTER_CLOUD_PACKAGE_PATH + PATH_IN_PACKAGE + "asset_package/module_with_assets.py:18"
    ),
    "little_richard": (
        DAGSTER_CLOUD_PACKAGE_PATH + PATH_IN_PACKAGE + "asset_package/__init__.py:4"
    ),
    "fats_domino": DAGSTER_CLOUD_PACKAGE_PATH + PATH_IN_PACKAGE + "asset_package/__init__.py:16",
    "miles_davis": (
        DAGSTER_CLOUD_PACKAGE_PATH
        + PATH_IN_PACKAGE
        + "asset_package/asset_subpackage/another_module_with_assets.py:6"
    ),
}


def test_link_code_references_to_git_if_cloud_no_cloud_context() -> None:
    from dagster_cloud_tests.metadata_tests import asset_package
    from dagster_cloud_tests.metadata_tests.asset_package import module_with_assets

    collection = load_assets_from_modules([asset_package, module_with_assets])
    collection_with_source_metadata = with_source_code_references(collection)
    # should be a no-op since we don't have the necessary environment variables set
    collection_with_potential_source_control_metadata = link_code_references_to_git_if_cloud(
        collection_with_source_metadata
    )

    for asset in collection_with_potential_source_control_metadata:
        if isinstance(asset, AssetsDefinition):
            op_name = asset.op.name
            assert op_name in EXPECTED_ORIGINS, f"Missing expected origin for op {op_name}"

            expected_file_path, expected_line_number = EXPECTED_ORIGINS[op_name].split(":")

            for key in asset.keys:
                assert "dagster/code_references" in asset.metadata_by_key[key]

                # `chuck_berry` is the only asset with source code metadata manually
                # attached to it, which coexists with the automatically attached metadata
                if op_name == "chuck_berry":
                    assert (
                        len(asset.metadata_by_key[key]["dagster/code_references"].code_references)
                        == 2
                    )
                else:
                    assert (
                        len(asset.metadata_by_key[key]["dagster/code_references"].code_references)
                        == 1
                    )

                assert isinstance(
                    asset.metadata_by_key[key]["dagster/code_references"].code_references[-1],
                    LocalFileCodeReference,
                )
                meta = cast(
                    "LocalFileCodeReference",
                    asset.metadata_by_key[key]["dagster/code_references"].code_references[-1],
                )

                assert meta.file_path == expected_file_path
                assert meta.line_number == int(expected_line_number)


# paramterize github gitlab
@pytest.mark.parametrize(
    "source_control_url, source_control_branch_base",
    [
        pytest.param(
            "https://github.com/dagster-io/dagster",
            "https://github.com/dagster-io/dagster/tree/master",
            id="github",
        ),
        pytest.param(
            "https://gitlab.com/dagster-io/dagster",
            "https://gitlab.com/dagster-io/dagster/-/tree/master",
            id="gitlab",
        ),
    ],
)
def test_link_code_references_to_git_if_cloud_cloud_context(
    source_control_url: str, source_control_branch_base: str
) -> None:
    from dagster_cloud_tests.metadata_tests import asset_package
    from dagster_cloud_tests.metadata_tests.asset_package import module_with_assets

    collection = load_assets_from_modules([asset_package, module_with_assets])
    collection_with_source_metadata = with_source_code_references(collection)

    # Setup the cloud context so that our code references are linked to source control
    with (
        environ(
            {
                "DAGSTER_CLOUD_DEPLOYMENT_NAME": "prod",
                "DAGSTER_CLOUD_GIT_URL": source_control_url,
                "DAGSTER_CLOUD_GIT_BRANCH": "master",
            }
        ),
        enter_loadable_target_origin_load_context(
            loadable_target_origin=LoadableTargetOrigin(python_file=__file__)
        ),
    ):
        # should link to source control since we have the necessary environment variables set
        collection_with_source_control_metadata = link_code_references_to_git_if_cloud(
            collection_with_source_metadata,
        )

        for asset in collection_with_source_control_metadata:
            if isinstance(asset, AssetsDefinition):
                op_name = asset.op.name
                assert op_name in EXPECTED_ORIGINS, f"Missing expected origin for op {op_name}"

                expected_file_path, expected_line_number = EXPECTED_ORIGINS[op_name].split(":")

                for key in asset.keys:
                    assert "dagster/code_references" in asset.metadata_by_key[key]

                    assert isinstance(
                        asset.metadata_by_key[key]["dagster/code_references"].code_references[-1],
                        UrlCodeReference,
                    )
                    meta = cast(
                        "UrlCodeReference",
                        asset.metadata_by_key[key]["dagster/code_references"].code_references[-1],
                    )

                    assert meta.url == (
                        f"{source_control_branch_base}/dagster-cloud/python_modules/dagster-cloud"
                        + (expected_file_path[len(DAGSTER_CLOUD_PACKAGE_PATH) :])
                        + f"#L{expected_line_number}"
                    )


def test_link_code_references_to_git_if_cloud_partial_cloud_context() -> None:
    from dagster_cloud_tests.metadata_tests import asset_package
    from dagster_cloud_tests.metadata_tests.asset_package import module_with_assets

    collection = load_assets_from_modules([asset_package, module_with_assets])
    collection_with_source_metadata = with_source_code_references(collection)

    # Setup the cloud context but do not provide the other necessary environment variables
    with (
        environ(
            {
                "DAGSTER_CLOUD_DEPLOYMENT_NAME": "prod",
            }
        ),
        enter_loadable_target_origin_load_context(
            loadable_target_origin=LoadableTargetOrigin(python_file=__file__)
        ),
    ):
        # We lack the branch/url environment variables so this should raise an error
        with pytest.raises(ValueError):
            link_code_references_to_git_if_cloud(
                collection_with_source_metadata,
            )


def test_link_code_references_to_git_if_cloud_override_cloud_context() -> None:
    from dagster_cloud_tests.metadata_tests import asset_package
    from dagster_cloud_tests.metadata_tests.asset_package import module_with_assets

    collection = load_assets_from_modules([asset_package, module_with_assets])
    collection_with_source_metadata = with_source_code_references(collection)

    # Setup the cloud context so that our code references are linked to source control
    with (
        environ({"DAGSTER_CLOUD_DEPLOYMENT_NAME": "prod"}),
        enter_loadable_target_origin_load_context(
            loadable_target_origin=LoadableTargetOrigin(python_file=__file__)
        ),
    ):
        # should link to source control despite missing the necessary environment variables
        # because we provide the necessary source control information as arguments
        collection_with_source_control_metadata = link_code_references_to_git_if_cloud(
            collection_with_source_metadata,
            git_url="https://github.com/dagster-io/other-repo",
            git_branch="main",
        )

        for asset in collection_with_source_control_metadata:
            if isinstance(asset, AssetsDefinition):
                op_name = asset.op.name
                assert op_name in EXPECTED_ORIGINS, f"Missing expected origin for op {op_name}"

                expected_file_path, expected_line_number = EXPECTED_ORIGINS[op_name].split(":")

                for key in asset.keys:
                    assert "dagster/code_references" in asset.metadata_by_key[key]

                    assert isinstance(
                        asset.metadata_by_key[key]["dagster/code_references"].code_references[-1],
                        UrlCodeReference,
                    )
                    meta = cast(
                        "UrlCodeReference",
                        asset.metadata_by_key[key]["dagster/code_references"].code_references[-1],
                    )

                    assert meta.url == (
                        "https://github.com/dagster-io/other-repo/tree/main/dagster-cloud/python_modules/dagster-cloud"
                        + (expected_file_path[len(DAGSTER_CLOUD_PACKAGE_PATH) :])
                        + f"#L{expected_line_number}"
                    )


def test_link_code_references_to_git_if_cloud_override_cloud_context_mapping() -> None:
    from dagster_cloud_tests.metadata_tests import asset_package
    from dagster_cloud_tests.metadata_tests.asset_package import module_with_assets

    collection = load_assets_from_modules([asset_package, module_with_assets])
    collection_with_source_metadata = with_source_code_references(collection)

    # Setup the cloud context so that our code references are linked to source control
    with (
        environ({"DAGSTER_CLOUD_DEPLOYMENT_NAME": "prod"}),
        enter_loadable_target_origin_load_context(
            loadable_target_origin=LoadableTargetOrigin(python_file=__file__)
        ),
    ):
        # should link to source control despite missing the necessary environment variables
        # because we provide the necessary source control information as arguments

        class _FilePathMapping(FilePathMapping):
            def convert_to_source_control_path(self, local_path: Path) -> str:
                return "override.py"

        collection_with_source_control_metadata = link_code_references_to_git_if_cloud(
            collection_with_source_metadata,
            git_url="https://github.com/dagster-io/other-repo",
            git_branch="main",
            file_path_mapping=_FilePathMapping(),
        )

        for asset in collection_with_source_control_metadata:
            if isinstance(asset, AssetsDefinition):
                op_name = asset.op.name
                assert op_name in EXPECTED_ORIGINS, f"Missing expected origin for op {op_name}"

                _expected_file_path, expected_line_number = EXPECTED_ORIGINS[op_name].split(":")

                for key in asset.keys:
                    assert "dagster/code_references" in asset.metadata_by_key[key]

                    assert isinstance(
                        asset.metadata_by_key[key]["dagster/code_references"].code_references[-1],
                        UrlCodeReference,
                    )
                    meta = cast(
                        "UrlCodeReference",
                        asset.metadata_by_key[key]["dagster/code_references"].code_references[-1],
                    )

                    assert meta.url.endswith(f"override.py#L{expected_line_number}")
